"""I automatically configure for use with cothread.catools

eg. attempt to use libca from this packages if available.

    try:
        import epicscorelibs.path.cothread
    except ImportError:
        pass
    from cothread.catools import caget
    ...
"""

import os, sys, warnings

from . import get_lib

if 'cothread.load_ca' in sys.modules or 'cothread.catools' in sys.modules:
    warnings.warn("epicscorelibs.path.cothread must be imported before cothread.catools to have effect")

# don't override if already set
if 'CATOOLS_LIBCA_PATH' not in os.environ:
    os.environ['CATOOLS_LIBCA_PATH'] = get_lib('ca')



"""Host configuration information using EPICS names and build flags.

Patterned after distutils.sysconfig
"""

from copy import deepcopy
import platform
import sysconfig
import distutils.sysconfig

__all__ = (
    'get_config_var',
    'get_config_vars',
)

def _makeconf():
    conf = {}

    # cf. Lib/platform.py in cpython source tree

    # map from python system name to epics OS_CLASS
    osname = conf['OS_CLASS'] = {
        'Linux': 'Linux',
        'Windows': 'WIN32',
        'Darwin': 'Darwin',
    }[platform.system()]

    # select epics CMPLR_CLASS
    CC = sysconfig.get_config_var('CC') or distutils.sysconfig.get_config_var('CC')
    if CC is None and osname=='WIN32':
        cmplrname = conf['CMPLR_CLASS'] = 'msvc'
    elif CC is None:
        raise RuntimeError("Unable to determine compiler")
    elif CC.find('gcc')!=-1:
        cmplrname = conf['CMPLR_CLASS'] = 'gcc'
    elif CC.find('clang')!=-1:
        cmplrname = conf['CMPLR_CLASS'] = 'clang'
    elif CC is None or CC.find('cl')!=-1:
        cmplrname = conf['CMPLR_CLASS'] = 'msvc'
    else:
        raise RuntimeError("Unable to identify compiler %s"%CC)

    machine = platform.machine().lower() # host CPU, may not reflect target bit width on windows
    bits = {
        '32bit':32,
        '64bit':64,
    }[platform.architecture()[0]]

    conf['POSIX'] = osname!='WIN32'

    # pick a host arch
    HA = None
    if osname=='Linux':
        if machine=='x86_64':
            HA = 'linux-x86_64'
        elif machine=='ppc':
            HA = 'linux-ppc'
        elif machine.startswith('arm') or machine=="aarch64":
            HA = 'linux-arm'
        elif machine.endswith('86'):
            HA = 'linux-x86'
        else:
            raise RuntimeError("Unsupported Linkage: " + machine)

    elif osname=='Darwin':
        if machine=='ppc':
            HA = 'darwin-ppc'
        else:
            HA = 'darwin-x86'

    elif osname=='WIN32':
        if bits==32 and cmplrname=='gcc':
            HA = 'win32-x86-mingw'
        elif bits==32 and cmplrname=='msvc':
            HA = 'win32-x86'
        elif bits==64 and cmplrname=='gcc':
            HA = 'windows-x64-mingw'
        elif bits==64 and cmplrname=='msvc':
            HA = 'windows-x64'

    if HA is None:
        raise RuntimeError("Unable to determine host arch")

    conf['EPICS_HOST_ARCH'] = conf['T_A'] = HA
    return conf

_config = _makeconf()

def _makebuild():
    build = {
        'CPPFLAGS':[],
        'CFLAGS':[],
        'CXXFLAGS':[],
        'LDFLAGS':[],
        'LDADD':[],
    }

    OS_CLASS = _config['OS_CLASS']

    if OS_CLASS=='Linux':
        build['CPPFLAGS'] += [('_GNU_SOURCE', None), ('_DEFAULT_SOURCE', None), ('linux', None)]
        # uncomment to force the linker to fail if any symbols are undefined.
        # helps to detect missing sources
        #build['LDFLAGS'] += ['-Wl,--no-undefined', '-Wl,--no-allow-shlib-undefined']
        build['LDADD'] += ['m', 'rt', 'dl']

        if not sysconfig.get_config_var('Py_DEBUG'):
            # don't include debug information in non-debug builds
            build['CFLAGS'] += ['-g0']
            build['CXXFLAGS'] += ['-g0']

    if OS_CLASS=='Darwin':
        build['CPPFLAGS'] += [('darwin', None)]
        build['CXXFLAGS'] += ['-mmacosx-version-min=10.7','-std=c++11', '-stdlib=libc++']
        build['LDFLAGS'] += ['-mmacosx-version-min=10.7','-std=c++11', '-stdlib=libc++']

    if OS_CLASS!='WIN32':
        build['CPPFLAGS'] += [('UNIX', None)]

    if OS_CLASS=='WIN32':
        build['CPPFLAGS'] += [('EPICS_BUILD_DLL', None), ('EPICS_CALL_DLL', None), ('NOMINMAX', None)]
        build['CXXFLAGS'] += ['-EHsc']
        build['LDADD'] += ['netapi32', 'ws2_32', 'advapi32', 'user32']

    return build

_config.update(_makebuild())

def get_config_var(key):
    return deepcopy(_config.get(key))

def get_config_vars():
    return deepcopy(_config)




from __future__ import print_function

import os
import ctypes

from .. import version

def test_version():
    print('Version:', version.version)
    print('Base Version:', version.base_version)
    print('ABI Requires:', version.abi_requires())




from Com cimport (epicsUInt8, epicsUInt16, epicsUInt32,
                  epicsInt8, epicsInt16, epicsInt32,
                  epicsFloat32, epicsFloat64,
                  epicsOldString,
                  epicsTimeStamp)

cdef extern from "<db_access.h>":
    enum: MAX_UNITS_SIZE
    enum: MAX_ENUM_STRING_SIZE
    enum: MAX_ENUM_STATES
    enum: DBF_STRING
    enum: DBF_INT
    enum: DBF_SHORT
    enum: DBF_FLOAT
    enum: DBF_ENUM
    enum: DBF_CHAR
    enum: DBF_LONG
    enum: DBF_DOUBLE
    enum: DBF_NO_ACCESS
    enum: LAST_TYPE

    enum: DBR_STRING
    enum: DBR_INT
    enum: DBR_SHORT
    enum: DBR_FLOAT
    enum: DBR_ENUM
    enum: DBR_CHAR
    enum: DBR_LONG
    enum: DBR_DOUBLE
    enum: DBR_STS_STRING
    enum: DBR_STS_SHORT
    enum: DBR_STS_INT
    enum: DBR_STS_FLOAT
    enum: DBR_STS_ENUM
    enum: DBR_STS_CHAR
    enum: DBR_STS_LONG
    enum: DBR_STS_DOUBLE
    enum: DBR_TIME_STRING
    enum: DBR_TIME_INT
    enum: DBR_TIME_SHORT
    enum: DBR_TIME_FLOAT
    enum: DBR_TIME_ENUM
    enum: DBR_TIME_CHAR
    enum: DBR_TIME_LONG
    enum: DBR_TIME_DOUBLE
    enum: DBR_GR_STRING
    enum: DBR_GR_SHORT
    enum: DBR_GR_INT
    enum: DBR_GR_FLOAT
    enum: DBR_GR_ENUM
    enum: DBR_GR_CHAR
    enum: DBR_GR_LONG
    enum: DBR_GR_DOUBLE
    enum: DBR_CTRL_STRING
    enum: DBR_CTRL_SHORT
    enum: DBR_CTRL_INT
    enum: DBR_CTRL_FLOAT
    enum: DBR_CTRL_ENUM
    enum: DBR_CTRL_CHAR
    enum: DBR_CTRL_LONG
    enum: DBR_CTRL_DOUBLE
    enum: DBR_PUT_ACKT
    enum: DBR_PUT_ACKS
    enum: DBR_STSACK_STRING
    enum: DBR_CLASS_NAME
    enum: LAST_BUFFER_TYPE

    # function type macros
    int VALID_DB_FIELD(int dbf)
    int INVALID_DB_FIELD(int dbf)
    int VALID_DB_REQ(int dbr)
    int INVALID_DB_REQ(int dbr)
    unsigned dbr_size_n(unsigned dbr, unsigned long count)
    void* dbr_value_ptr(void* buf, int dbr)
    int dbf_type_is_valid(int dbf)
    int dbr_type_is_valid(int dbr)
    int dbr_type_is_plain(int dbr)
    int dbr_type_is_STS(int dbr)
    int dbr_type_is_TIME(int dbr)
    int dbr_type_is_GR(int dbr)
    int dbr_type_is_CTRL(int dbr)
    int dbr_type_is_STRING(int dbr)
    int dbr_type_is_SHORT(int dbr)
    int dbr_type_is_FLOAT(int dbr)
    int dbr_type_is_ENUM(int dbr)
    int dbr_type_is_CHAR(int dbr)
    int dbr_type_is_LONG(int dbr)
    int dbr_type_is_DOUBLE(int dbr)
    const char* dbf_type_to_text(int dbf)
    const char* dbr_type_to_text(int dbr)
    int dbf_type_to_DBR(int dbr)
    int dbf_type_to_DBR_STS(int dbr)
    int dbf_type_to_DBR_TIME(int dbr)
    int dbf_type_to_DBR_GR(int dbr)
    int dbf_type_to_DBR_CTRL(int dbr)

    ctypedef epicsOldString dbr_string_t
    ctypedef epicsUInt8 dbr_char_t
    ctypedef epicsInt16 dbr_short_t
    ctypedef epicsUInt16 dbr_ushort_t
    ctypedef epicsInt16 dbr_int_t
    ctypedef epicsUInt16 dbr_enum_t
    ctypedef epicsInt32 dbr_long_t
    ctypedef epicsUInt32 dbr_ulong_t
    ctypedef epicsFloat32 dbr_float_t
    ctypedef epicsFloat64 dbr_double_t
    ctypedef epicsUInt16 dbr_put_ackt_t
    ctypedef epicsUInt16 dbr_put_acks_t
    ctypedef epicsOldString dbr_stsack_string_t
    ctypedef epicsOldString dbr_class_name_t

    cdef struct dbr_sts_string:
        dbr_short_t status
        dbr_short_t severity
        dbr_string_t value

    cdef struct dbr_stsack_string:
        dbr_ushort_t	status
        dbr_ushort_t	severity
        dbr_ushort_t	ackt
        dbr_ushort_t	acks
        dbr_string_t	value

    cdef struct dbr_sts_int:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	value

    cdef struct dbr_sts_short:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	value

    cdef struct dbr_sts_float:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_float_t	value

    cdef struct dbr_sts_enum:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_enum_t	value

    cdef struct dbr_sts_char:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_char_t	RISC_pad
        dbr_char_t	value

    cdef struct dbr_sts_long:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_long_t	value

    cdef struct dbr_sts_double:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_long_t	RISC_pad
        dbr_double_t	value


    # VALUES WITH STATUS AND TIME STRUCTURES

    cdef struct dbr_time_string:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_string_t	value

    cdef struct dbr_time_short:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_short_t	RISC_pad
        dbr_short_t	value

    cdef struct dbr_time_float:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_float_t	value

    cdef struct dbr_time_enum:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_short_t	RISC_pad
        dbr_enum_t	value

    cdef struct dbr_time_char:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_short_t	RISC_pad0
        dbr_char_t	RISC_pad1
        dbr_char_t	value

    cdef struct dbr_time_long:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_long_t	value

    cdef struct dbr_time_double:
        dbr_short_t	status
        dbr_short_t	severity
        epicsTimeStamp	stamp
        dbr_long_t	RISC_pad
        dbr_double_t	value

    # VALUES WITH STATUS AND GRAPHIC STRUCTURES

    cdef struct dbr_gr_int:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_short_t	upper_disp_limit
        dbr_short_t	lower_disp_limit
        dbr_short_t	upper_alarm_limit
        dbr_short_t	upper_warning_limit
        dbr_short_t	lower_warning_limit
        dbr_short_t	lower_alarm_limit
        dbr_short_t	value

    cdef struct dbr_gr_short:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_short_t	upper_disp_limit
        dbr_short_t	lower_disp_limit
        dbr_short_t	upper_alarm_limit
        dbr_short_t	upper_warning_limit
        dbr_short_t	lower_warning_limit
        dbr_short_t	lower_alarm_limit
        dbr_short_t	value

    cdef struct dbr_gr_float:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	precision
        dbr_short_t	RISC_pad0
        char units[MAX_UNITS_SIZE]
        dbr_float_t	upper_disp_limit
        dbr_float_t	lower_disp_limit
        dbr_float_t	upper_alarm_limit
        dbr_float_t	upper_warning_limit
        dbr_float_t	lower_warning_limit
        dbr_float_t	lower_alarm_limit
        dbr_float_t	value

    cdef struct dbr_gr_enum:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	no_str
        char strs[MAX_ENUM_STATES][MAX_ENUM_STRING_SIZE]
        dbr_enum_t	value

    cdef struct dbr_gr_char:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_char_t	upper_disp_limit
        dbr_char_t	lower_disp_limit
        dbr_char_t	upper_alarm_limit
        dbr_char_t	upper_warning_limit
        dbr_char_t	lower_warning_limit
        dbr_char_t	lower_alarm_limit
        dbr_char_t	RISC_pad
        dbr_char_t	value

    cdef struct dbr_gr_long:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_long_t	upper_disp_limit
        dbr_long_t	lower_disp_limit
        dbr_long_t	upper_alarm_limit
        dbr_long_t	upper_warning_limit
        dbr_long_t	lower_warning_limit
        dbr_long_t	lower_alarm_limit
        dbr_long_t	value

    cdef struct dbr_gr_double:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	precision
        dbr_short_t	RISC_pad0
        char units[MAX_UNITS_SIZE]
        dbr_double_t	upper_disp_limit
        dbr_double_t	lower_disp_limit
        dbr_double_t	upper_alarm_limit
        dbr_double_t	upper_warning_limit
        dbr_double_t	lower_warning_limit
        dbr_double_t	lower_alarm_limit
        dbr_double_t	value

    # VALUES WITH STATUS, GRAPHIC and CONTROL STRUCTURES */

    cdef struct dbr_ctrl_int:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_short_t	upper_disp_limit
        dbr_short_t	lower_disp_limit
        dbr_short_t	upper_alarm_limit
        dbr_short_t	upper_warning_limit
        dbr_short_t	lower_warning_limit
        dbr_short_t	lower_alarm_limit
        dbr_short_t	upper_ctrl_limit
        dbr_short_t	lower_ctrl_limit
        dbr_short_t	value

    cdef struct dbr_ctrl_short:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_short_t	upper_disp_limit
        dbr_short_t	lower_disp_limit
        dbr_short_t	upper_alarm_limit
        dbr_short_t	upper_warning_limit
        dbr_short_t	lower_warning_limit
        dbr_short_t	lower_alarm_limit
        dbr_short_t	upper_ctrl_limit
        dbr_short_t	lower_ctrl_limit
        dbr_short_t	value

    cdef struct dbr_ctrl_float:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	precision
        dbr_short_t	RISC_pad
        char units[MAX_UNITS_SIZE]
        dbr_float_t	upper_disp_limit
        dbr_float_t	lower_disp_limit
        dbr_float_t	upper_alarm_limit
        dbr_float_t	upper_warning_limit
        dbr_float_t	lower_warning_limit
        dbr_float_t	lower_alarm_limit
        dbr_float_t	upper_ctrl_limit
        dbr_float_t	lower_ctrl_limit
        dbr_float_t	value

    cdef struct dbr_ctrl_enum:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	no_str
        char strs[MAX_ENUM_STATES][MAX_ENUM_STRING_SIZE]
        dbr_enum_t	value

    cdef struct dbr_ctrl_char:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_char_t	upper_disp_limit
        dbr_char_t	lower_disp_limit
        dbr_char_t	upper_alarm_limit
        dbr_char_t	upper_warning_limit
        dbr_char_t	lower_warning_limit
        dbr_char_t	lower_alarm_limit
        dbr_char_t	upper_ctrl_limit
        dbr_char_t	lower_ctrl_limit
        dbr_char_t	RISC_pad
        dbr_char_t	value

    cdef struct dbr_ctrl_long:
        dbr_short_t	status
        dbr_short_t	severity
        char units[MAX_UNITS_SIZE]
        dbr_long_t	upper_disp_limit
        dbr_long_t	lower_disp_limit
        dbr_long_t	upper_alarm_limit
        dbr_long_t	upper_warning_limit
        dbr_long_t	lower_warning_limit
        dbr_long_t	lower_alarm_limit
        dbr_long_t	upper_ctrl_limit
        dbr_long_t	lower_ctrl_limit
        dbr_long_t	value

    cdef struct dbr_ctrl_double:
        dbr_short_t	status
        dbr_short_t	severity
        dbr_short_t	precision
        dbr_short_t	RISC_pad0
        char units[MAX_UNITS_SIZE]
        dbr_double_t	upper_disp_limit
        dbr_double_t	lower_disp_limit
        dbr_double_t	upper_alarm_limit
        dbr_double_t	upper_warning_limit
        dbr_double_t	lower_warning_limit
        dbr_double_t	lower_alarm_limit
        dbr_double_t	upper_ctrl_limit
        dbr_double_t	lower_ctrl_limit
        dbr_double_t	value

cdef extern from "<cadef.h>":
    enum: CA_OP_GET
    enum: CA_OP_PUT
    enum: CA_OP_CREATE_CHANNEL
    enum: CA_OP_ADD_EVENT
    enum: CA_OP_CLEAR_EVENT
    enum: CA_OP_OTHER
    enum: CA_OP_CONN_UP
    enum: CA_OP_CONN_DOWN
    enum: TYPENOTCONN

    cdef struct oldChannelNotify:
        pass # opaque
    cdef struct oldSubscription:
        pass # opaque

    ctypedef oldChannelNotify* chid
    ctypedef chid                    chanId
    ctypedef long                    chtype
    ctypedef oldSubscription * evid
    ctypedef double                  ca_real
    ctypedef unsigned capri

    cdef struct connection_handler_args:
        chanId chid
        long op
    ctypedef void caCh (connection_handler_args args)

    cdef struct ca_access_rights:
        unsigned read_access
        unsigned write_access
    ctypedef access_rights_handler_args caar

    cdef struct access_rights_handler_args:
        chanId chid
        caar ar
    ctypedef void caArh (access_rights_handler_args args)

    cdef struct event_handler_args:
        void            *usr
        chanId          chid
        long            type
        long            count
        const void      *dbr
        int             status
    ctypedef void caEventCallBackFunc (event_handler_args)

    cdef struct exception_handler_args:
        void            *usr
        chanId          chid
        long            type
        long            count
        void            *addr
        long            stat
        long            op
        const char      *ctx
        const char      *pFile
        unsigned        lineNo
    ctypedef void caExceptionHandler (exception_handler_args)

    cdef struct ca_client_context:
        pass # opaque

    short ca_field_type (chid chan)
    unsigned long ca_element_count (chid chan)
    const char * ca_name (chid chan)
    void ca_set_puser (chid chan, void *puser)
    void * ca_puser (chid chan)
    unsigned ca_read_access (chid chan)
    unsigned ca_write_access (chid chan)

    cdef enum channel_state:
        cs_never_conn
        cs_prev_conn
        cs_conn
        cs_closed

    channel_state ca_state (chid chan)

    int ca_task_initialize ()

    cdef enum ca_preemptive_callback_select:
        ca_disable_preemptive_callback
        ca_enable_preemptive_callback

    int ca_context_create (ca_preemptive_callback_select select)
    void ca_detach_context () 

    int ca_task_exit ()
    void ca_context_destroy ()

    int ca_create_channel(const char *pChanName, caCh *pConnStateCallback, void *pUserPrivate, capri priority, chid *pChanID)

    int ca_change_connection_event(chid chan, caCh * pfunc)

    int ca_replace_access_rights_event (chid chan, caArh *pfunc)

    int ca_add_exception_event(caExceptionHandler *pfunc, void *pArg)

    int ca_clear_channel(chid chanId)

    int ca_array_put(chtype type, unsigned long count, chid chanId, const void * pValue)

    int ca_array_put_callback(chtype type, unsigned long count, chid chanId, const void * pValue, caEventCallBackFunc * pFunc, void * pArg)

    int ca_array_get( chtype type, unsigned long count, chid chanId, void * pValue)

    int ca_array_get_callback( chtype type, unsigned long count, chid chanId, caEventCallBackFunc * pFunc, void * pArg)

    int ca_create_subscription( chtype type, unsigned long count, chid chanId, long mask, caEventCallBackFunc * pFunc, void * pArg, evid * pEventID)

    int ca_clear_subscription( evid eventID)

    chid ca_evid_to_chid ( evid id )

    int ca_pend_event (ca_real timeOut)
    int ca_poll()

    int ca_pend_io (ca_real timeOut)

    int ca_pend (ca_real timeout, int early)

    int ca_test_io ()

    int ca_flush_io ()

    const char * ca_host_name (chid channel)
    # thread safe version
    unsigned ca_get_host_name ( chid pChan, char *pBuf, unsigned bufLength )

    int ca_v42_ok (chid chan)

    const char* ca_version ()

    ca_client_context* ca_current_context ()
    int ca_attach_context ( ca_client_context * context )

    int ca_client_status ( unsigned level )
    int ca_context_status ( ca_client_context *, unsigned level )


    int ca_modify_user_name ( const char *pUserName )
    int ca_modify_host_name ( const char *pHostName )




cdef extern from "<epicsTypes.h>":
    enum: MAX_STRING_SIZE

    ctypedef char            epicsOldString[MAX_STRING_SIZE]

    ctypedef char            epicsInt8
    ctypedef unsigned char   epicsUInt8
    ctypedef short           epicsInt16
    ctypedef unsigned short  epicsUInt16
    ctypedef int             epicsInt32
    ctypedef unsigned int    epicsUInt32
    ctypedef long long       epicsInt64
    ctypedef unsigned long long epicsUInt64

    ctypedef epicsUInt16     epicsEnum16
    ctypedef float           epicsFloat32
    ctypedef double          epicsFloat64
    ctypedef epicsInt32      epicsStatus

cdef extern from "<epicsTime.h>" nogil:
    enum: POSIX_TIME_AT_EPICS_EPOCH

    cdef struct epicsTimeStamp:
        epicsUInt32    secPastEpoch   # seconds since 0000 Jan 1, 1990
        epicsUInt32    nsec

    int epicsTimeGetCurrent ( epicsTimeStamp * pDest )

cdef extern from "<errSymTbl.h>" nogil:
    const char* errSymMsg(long status)
    void errSymLookup(long status, char *pBuf, size_t bufLength)

cdef extern from "<errlog.h>" nogil:
    int eltc(int yesno)
    int errlogPrintf(const char *pformat, ...)
    int errlogPrintfNoConsole(const char *pformat, ...)

cdef extern from "<iocsh.h>" nogil:
    int iocsh(const char *pathname)
    int iocshCmd(const char *cmd)
    int iocshLoad(const char *pathname, const char* macros)
    int iocshRun(const char *cmd, const char* macros)

cdef extern from "<iocshRegisterCommon.h>" nogil:
    void iocshRegisterCommon()

cdef extern from "<epicsFindSymbol.h>" nogil:
    void* epicsLoadLibrary(const char *name);
    const char* epicsLoadError()
    void* epicsFindSymbol(const char *name)

cdef extern from "<epicsExit.h>" nogil:
    ctypedef void (*epicsExitFunc)(void *arg)

    void epicsExit(int status)
    void epicsExitLater(int status)
    void epicsExitCallAtExits()
    int epicsAtExit(epicsExitFunc func, void *arg)
    int epicsAtExit3(epicsExitFunc func, void *arg, const char* name)

    void epicsExitCallAtThreadExits()
    int epicsAtThreadExit(epicsExitFunc func, void *arg)

cdef extern from "<initHooks.h>" nogil:
    ctypedef int initHookState

    ctypedef void (*initHookFunction)(initHookState state)
    int initHookRegister(initHookFunction func)
    const char *initHookName(int state)

cdef extern from "<alarm.h>" nogil:
    enum: NO_ALARM
    enum: MINOR_ALARM
    enum: MAJOR_ALARM
    enum: INVALID_ALARM
    enum: ALARM_NSEV

    enum: READ_ALARM
    enum: WRITE_ALARM
    enum: HIHI_ALARM
    enum: HIGH_ALARM
    enum: LOLO_ALARM
    enum: LOW_ALARM
    enum: STATE_ALARM
    enum: COS_ALARM
    enum: COMM_ALARM
    enum: TIMEOUT_ALARM
    enum: HW_LIMIT_ALARM
    enum: CALC_ALARM
    enum: SCAN_ALARM
    enum: LINK_ALARM
    enum: SOFT_ALARM
    enum: BAD_SUB_ALARM
    enum: UDF_ALARM
    enum: DISABLE_ALARM
    enum: SIMM_ALARM
    enum: READ_ACCESS_ALARM
    enum: WRITE_ACCESS_ALARM
    enum: ALARM_NSTATUS

    extern const char *epicsAlarmSeverityStrings [ALARM_NSEV]
    extern const char *epicsAlarmConditionStrings [ALARM_NSTATUS]

cdef extern from "<epicsMutex.h>" nogil:
    struct epicsMutexParm:
        pass
    ctypedef epicsMutexParm* epicsMutexId
    enum epicsMutexLockStatus_t:
        epicsMutexLockOK
        epicsMutexLockTimeout
        epicsMutexLockError
    ctypedef epicsMutexLockStatus_t epicsMutexLockStatus

    epicsMutexId epicsMutexCreate()
    void epicsMutexDestroy(epicsMutexId)
    void epicsMutexUnlock(epicsMutexId)
    epicsMutexLockStatus epicsMutexLock(epicsMutexId)
    epicsMutexLockStatus epicsMutexTryLock(epicsMutexId)




from __future__ import print_function

import os
import ctypes

from .. import path

def test_loading():
    print("PATH", os.environ.get('PATH'))
    print("LD_LIBRARY_PATH", os.environ.get('LD_LIBRARY_PATH'))
    print("DYLD_LIBRARY_PATH", os.environ.get('DYLD_LIBRARY_PATH'))
    print(path.get_lib('ca'))

    ca = ctypes.CDLL(path.get_lib('ca'), mode=ctypes.RTLD_GLOBAL)

    if path.OS_CLASS=='WIN32':
        # Base libs contain functions using a mixture of cdecl and stdcall calling conventions
        # make WinDLL w/o actually loading again
        ca_s = ctypes.WinDLL(ca._name, handle=ca._handle)
        # functions from dependent DLLs aren't accessible through this handle
        Com = ctypes.CDLL(path.get_lib('Com'), mode=ctypes.RTLD_GLOBAL)
        Com_s = ctypes.WinDLL(Com._name, handle=Com._handle)
    else:
        Com = Com_s = ca_s = ca

    assert ca._handle==ca_s._handle, (ca._handle, ca_s._handle)

    # from libCom
    epicsThreadGetCPUs = Com.epicsThreadGetCPUs
    epicsThreadGetNameSelf = Com_s.epicsThreadGetNameSelf
    epicsThreadGetNameSelf.restype = ctypes.c_char_p
    # from libca
    ca_version = ca_s.ca_version
    ca_version.restype = ctypes.c_char_p

    print("epicsThreadGetCPUs()", epicsThreadGetCPUs())
    print("epicsThreadGetNameSelf()", epicsThreadGetNameSelf())
    print("ca_version()", ca_version())

    # test if re-loading is actually re-loading

    Com2 = ctypes.CDLL(path.get_lib('Com'), mode=ctypes.RTLD_GLOBAL)

    addrA = ctypes.c_char_p.in_dll(Com, 'EPICS_CA_ADDR_LIST')
    addrB = ctypes.c_char_p.in_dll(Com2, 'EPICS_CA_ADDR_LIST')
    print(addrA, addrB)

    assert ctypes.addressof(addrA)==ctypes.addressof(addrB), (addrA, addrB)


def test_ca_loading():
    from epicscorelibs.ca import cadef, dbr
    assert cadef
    assert dbr




from . cimport Com

cdef extern from "<iocInit.h>" nogil:
    int iocInit()
    int iocBuild()
    int iocBuildIsolated()
    int iocRun()
    int iocPause()
    int iocShutdown()

cdef extern from "<dbUnitTest.h>" nogil:
    void testdbPrepare()
    void testdbReadDatabase(const char* file, const char* path, const char* substitutions)
    void testIocInitOk()
    void testIocShutdownOk()
    void testdbCleanup()

cdef extern from "<dbStaticLib.h>" nogil:
    struct DBBASE:
        pass
    struct dbRecordNode:
        void* precord
        char* recordname
    struct dbFldDes:
        char* name
        int field_type # dbfType
        short special
        short size
        unsigned short offset
    struct DBENTRY "dbEntry":
        dbRecordNode* precnode
        dbFldDes     *pflddes

    DBBASE* pdbbase

    void dbInitEntry(DBBASE *pdbbase, DBENTRY *pdbentry)
    void dbFinishEntry(DBENTRY *pdbentry)
    long dbFindRecord(DBENTRY *pdbentry, const char *pname)
    long dbFindField(DBENTRY *pdbentry, const char *pfieldName)
    long dbEntryToAddr(const DBENTRY *pdbentry, dbAddr *paddr)

cdef extern from "<dbAccess.h>" nogil:
    int dbLoadDatabase(const char *filename, const char *path, const char *substitutions)
    int dbLoadRecords (const char* filename,                   const char* substitutions)

    long dbNameToAddr(const char *pname, dbAddr *paddr)

    long dbGetField(dbAddr *paddr, short dbrType, void *pbuffer, long *options, long *nRequest,void *pfl)
    long dbGet     (dbAddr *paddr, short dbrType, void *pbuffer, long *options, long *nRequest,void *pfl)
    long dbPutField(dbAddr *,short dbrType,const void *pbuffer,long nRequest)
    long dbPut     (dbAddr *,short dbrType,const void *pbuffer,long nRequest)
    long dbBufferSize(short dbrType,long options,long nRequest)
    long dbValueSize(short dbrType)

    long dbProcess(dbCommon *precord)

    void dbInitEntryFromRecord(dbCommon* prec, DBENTRY* pdbentry)

cdef extern from "<dbLoadTemplate.h>" nogil:
    int dbLoadTemplate(const char *sub_file, const char *cmd_collect)

cdef extern from "<dbCommon.h>" nogil:
    struct dbCommon:
        const char* const  name
        Com.epicsTimeStamp time
        Com.epicsUInt8     pact
        void*              dpvt
        # ... and sooo many more

cdef extern from "<dbLock.h>" nogil:
    void dbScanLock(dbCommon *prec)
    void dbScanUnlock(dbCommon *prec)

cdef extern from "<dbAddr.h>" nogil:
    struct dbAddr:
        dbCommon* precord
        void* pfield
        dbFldDes *pfldDes
        long    no_elements
        short   field_type
        short   field_size
        short   special
        short   dbr_field_type

cdef extern from "<dbChannel.h>" nogil:
    struct dbChannel:
        const char *name
        dbAddr addr
        long  final_no_elements
        short final_field_size
        short final_type
        # otherwise opaque

    const char* dbChannelName(dbChannel*)
    dbCommon* dbChannelRecord(dbChannel*)
    dbFldDes* dbChannelFldDes(dbChannel*)
    long dbChannelElements(dbChannel*)
    short dbChannelFieldType(dbChannel*)
    short dbChannelExportType(dbChannel*)
    short dbChannelExportCAType(dbChannel*)
    short dbChannelFieldSize(dbChannel*)
    long dbChannelFinalElements(dbChannel*)
    short dbChannelFinalFieldType(dbChannel*)
    short dbChannelFinalCAType(dbChannel*)
    short dbChannelFinalFieldSize(dbChannel*)
    short dbChannelSpecial(dbChannel*)
    void* dbChannelField(dbChannel*)

cdef extern from "<registryFunction.h>" nogil:
    ctypedef void (*REGISTRYFUNCTION)()
    int registryFunctionAdd(const char *name, REGISTRYFUNCTION func)

cdef extern from "<registryCommon.h>" nogil:
    #void registerRecordTypes(DBBASE *pbase, int nRecordTypes, const char * const *recordTypeNames, const recordTypeLocation *rtl)
    void registerDevices(DBBASE *pbase, int nDevices, const char * const *deviceSupportNames, const dset * const *devsl)
    #void registerDrivers(DBBASE *pbase, int nDrivers, const char * const *driverSupportNames, struct drvet * const *drvsl)
    #void registerJLinks(DBBASE *pbase, int nDrivers, jlif * const *jlifsl);


cdef extern from "<recGbl.h>" nogil:
    int recGblSetSevr(void *precord, Com.epicsEnum16 new_stat, Com.epicsEnum16 new_sevr)
    void recGblGetTimeStamp(void *precord)

cdef extern from "<dbScan.h>" nogil:
    ctypedef void* EVENTPVT
    EVENTPVT eventNameToHandle(const char* event)
    void postEvent(EVENTPVT epvt)

    struct ioscan_head:
        pass
    ctypedef ioscan_head* IOSCANPVT
    void scanIoInit(IOSCANPVT *ppios)
    unsigned int scanIoRequest(IOSCANPVT pios)
    unsigned int scanIoImmediate(IOSCANPVT pios, int prio)
    ctypedef void (*io_scan_complete)(void *usr, IOSCANPVT, int prio)
    void scanIoSetComplete(IOSCANPVT, io_scan_complete, void *usr)

cdef extern from "<dbFldTypes.h>" nogil:
    enum: DBF_STRING
    enum: DBF_CHAR
    enum: DBF_UCHAR
    enum: DBF_SHORT
    enum: DBF_USHORT
    enum: DBF_LONG
    enum: DBF_ULONG
    enum: DBF_INT64
    enum: DBF_UINT64
    enum: DBF_FLOAT
    enum: DBF_DOUBLE
    enum: DBF_ENUM
    enum: DBF_MENU
    enum: DBF_DEVICE
    enum: DBF_INLINK
    enum: DBF_OUTLINK
    enum: DBF_FWDLINK
    enum: DBF_NOACCESS

cdef extern from "<link.h>" nogil:
    struct instio:
        char *string

    union value:
        # ignore
        instio instio

    struct link:
        dbCommon *precord
        short type
        short flags
        # lset *lset
        char *text
        value value

    ctypedef link DBLINK

cdef extern from "<devSup.h>" nogil:
    enum: CONSTANT
    enum: INST_IO

    struct dset:
        pass # may be a typedef of 'typed_dset'
    struct typed_dset:
        long number
        long (*report)(int lvl)
        long (*init)(int after)
        long (*init_record)(dbCommon *prec)
        long (*get_ioint_info)(int detach, dbCommon *prec, IOSCANPVT* pscan)

    struct dsxt:
        long (*add_record)(dbCommon *precord)
        long (*del_record)(dbCommon *precord)

    ctypedef long (*DEVSUPFUN)()

    link* dbGetDevLink(dbCommon* prec)
    void devExtend(dsxt *pdsxt)

cdef extern from "<dbNotify.h>" nogil:
    enum notifyRequestType:
        processRequest
        putProcessRequest
        processGetRequest
        putProcessGetRequest
    enum _notifyPutType:
        putDisabledType
        putFieldType
        putType
    ctypedef _notifyPutType notifyPutType
    enum _notifyGetType:
        getFieldType
        getType
    ctypedef _notifyGetType notifyGetType
    struct processNotify:
        notifyRequestType requestType
        dbChannel *chan
        int              (*putCallback)(processNotify *,notifyPutType type)
        void             (*getCallback)(processNotify *,notifyGetType type)
        void             (*doneCallback)(processNotify *)
        void             *usrPvt

    void dbProcessNotify(processNotify *pprocessNotify)
    void dbNotifyCancel(processNotify *pprocessNotify)



'''Minimal softIoc
'''
from __future__ import print_function

import ctypes
import code
import argparse
import sys
import os

from . import path


# The libraries we need to run up a soft IOC, don't load dbCore twice
if path.OS_CLASS == "WIN32":
    Com = ctypes.WinDLL(path.get_lib("Com"), mode=ctypes.RTLD_GLOBAL)
else:
    Com = ctypes.CDLL(path.get_lib("Com"), mode=ctypes.RTLD_GLOBAL)
dbCore = ctypes.CDLL(path.get_lib("dbCore"), mode=ctypes.RTLD_GLOBAL)
dbRecStd = ctypes.CDLL(path.get_lib("dbRecStd"), mode=ctypes.RTLD_GLOBAL)

# The functions we need from those libraries

iocshCmd = Com.iocshCmd
iocshCmd.restype = ctypes.c_int
iocshCmd.argtypes = [ctypes.c_char_p]

iocshRegisterCommon = dbCore.iocshRegisterCommon
iocshRegisterCommon.restype = None
iocshRegisterCommon.argtypes = []

# dbLoadDatabase(file, path, subs)
dbLoadDatabase = dbCore.dbLoadDatabase
dbLoadDatabase.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]

# pdbbase is valid after the DBD is loaded
pdbbase = ctypes.c_void_p.in_dll(dbCore, "pdbbase")

# softIoc_registerRecordDeviceDriver(pdbbase)
#registerRecordDeviceDriver = dbRecStd.softIoc_registerRecordDeviceDriver
registerRecordDeviceDriver = dbCore.registerAllRecordDeviceDrivers
registerRecordDeviceDriver.argtypes = [ctypes.c_void_p]

# dbLoadRecords(file, subs)
dbLoadRecords = dbCore.dbLoadRecords
dbLoadRecords.argtypes = [ctypes.c_char_p, ctypes.c_char_p]

# iocInit()
iocInit = dbCore.iocInit

def ioc(cmd):
    '''Execute IOC shell command
    '''
    return iocshCmd(cmd.encode())

def start_ioc(database=None, macros='', dbs=None):
    if dbs is None:
        dbs = []
    if database is not None:
        dbs += [(database, macros)]

    def out(msg, *args):
        sys.stderr.write(msg%args)
        sys.stderr.flush()

    iocshRegisterCommon()

    out('IOC Starting w/ %s \n', dbs)
    dbd = os.path.join(path.base_path, "dbd")
    if dbLoadDatabase(b"base.dbd", dbd.encode(), None):
        raise RuntimeError('Error loading '+dbd)

    out('IOC dbd loaded\n')
    if registerRecordDeviceDriver(pdbbase):
        raise RuntimeError('Error registering')

    out('IOC rRDD\n')
    for database, macros in dbs:
        if dbLoadRecords(database.encode(), macros.encode()):
            raise RuntimeError('Error loading %s with %s'%(database, macros))
        out('IOC db loaded\n')
    if iocInit():
        raise RuntimeError('Error starting IOC')
    out('IOC Running\n')


def main():
    class DbAction(argparse.Action):
        def __call__(self, parser, ns, values, opt):
            ns.database.append((values, ns.macros))
    parser = argparse.ArgumentParser(
        description="Run a cut down softIoc by calling the functions from Python")
    parser.add_argument('-m', '--macros', default="",
        help="Macro definitions when expanding database, e.g. -d \"macro=value,macro2=value2\"")
    parser.add_argument('-d', '--database', default=[], action=DbAction,
        help="Path to database file to load")
    args = parser.parse_args()
    start_ioc(dbs=args.database)
    code.interact(local={
        'exit':sys.exit,
        'ioc':ioc,
    })


# Run with python -m epicscorelibs.ioc -d /path/to/db -m macro=value,macro2=value2
if __name__ == "__main__":
    main()



import subprocess
import os
import sys
import time
import ctypes
import tempfile
import random
import string
import unittest

try:
    from queue import Queue
except ImportError:
    from Queue import Queue

from epicscorelibs.ca import cadef, dbr


class Channel:
    def __init__(self, name):
        self.name = name
        self.chid = ctypes.c_void_p()

    @property
    def _as_parameter_(self):
        # Setting this allows a channel object to autoconvert into the chid
        # when passed to ca_ functions.
        return self.chid.value


def test_running_ioc():
    if os.environ.get("PYTAG", "") == "win_amd64" \
            and sys.version_info[:2] == (3, 4):
        # This particular combination is broken. All the callbacks seem to
        # return junk. Not sure why
        raise unittest.SkipTest()

    # Get something reasonably unique for the PV prefix
    pv = ''.join(random.choice(string.ascii_uppercase) for _ in range(12))
    dbfile = tempfile.NamedTemporaryFile(delete=False)
    dbfile.write(("""
record(stringin, "%s") {
    field("VAL", "blah")
}
""" % pv).encode())
    dbfile.close()

    try:
        # Start the IOC
        proc = subprocess.Popen(
            [sys.executable, '-m', 'epicscorelibs.ioc', '-d', dbfile.name],
            stdin=subprocess.PIPE,
        )

        # Effectively do a caget with these parameters
        datatype = dbr.DBR_STRING
        format = dbr.FORMAT_RAW
        count = 0
        results = Queue()

        # Polling mode, different from what cothread/aioca does
        cadef.ca_context_create(0)

        # Create a channel to our record
        @cadef.connection_handler
        def on_connect(args):
            print("on_connect got op %s" % args.op)
            results.put(args.op)

        channel = Channel(pv)
        cadef.ca_create_channel(
            pv, on_connect, None, 0, ctypes.byref(channel.chid))
        # Check that it connected in 5s
        cadef.ca_pend_event(5.0)
        assert results.get(timeout=1.0) == cadef.CA_OP_CONN_UP
        dbrcode, dbr_to_value = dbr.type_to_dbr(channel, datatype, format)

        # Get the value
        @cadef.event_handler
        def event_handler(args):
            print("event_handler got dbr %s" % args.raw_dbr)
            results.put(dbr_to_value(args.raw_dbr, args.type, args.count))

        cadef.ca_array_get_callback(dbrcode, count, channel, event_handler, None)

        # Check that we get result in 1s
        cadef.ca_pend_event(1.0)
        assert results.get(timeout=1.0) == "blah"

    finally:
        print('close IOC stdin')
        proc.stdin.close()
        # Popen.wait(timeout=) added in 3.3
        for n in range(10):
            ret = proc.poll()
            if ret is not None:
                print('IOC exits with',ret)
                break
            time.sleep(1.0)
        else:
            print('IOC killing')
            proc.terminate()
            proc.wait()
            print('IOC dead')
        os.unlink(dbfile.name)



"""Version information for the epicscorelibs package.

Version numbers are encoded as:

BASEVERSION.99.ABI.SRC

"""
import re
from pkg_resources import get_distribution, parse_version

__all__ = (
    'version',
    'base_version',
    'abi_requires',
)

_me = get_distribution('epicscorelibs')

version = _me.version # as a string

# in old setuptools parse_version() returns a tuple
#version_info = parse_version(version)
#base_version = version_info.base_version # w/o pre-release suffix

base_version, = re.match(r'([\d.]+)(?:[a-z]+\d+)?', version).groups()

def abi_requires():
    """Return a version requirement string which identifies
    a range of version which will be ABI compatible with this one.
    For use by modules with non-python dependencies on our libraries.

    eg. "epicscorelibs >=7.0.1.99.1.1a5, <7.0.1.99.2"
    """
    parts = base_version.split('.')
    assert parts[-3]=='99', version
    abi = int(parts[-2])

    return 'epicscorelibs >=%s, <%s.%d'%(version, '.'.join(parts[:-2]), abi+1)





import os
from glob import glob

from ..config import get_config_var

__all__ = (
    'include_path',
    'lib_path',
    'get_lib',
)

base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))


# parts of library file names
OS_CLASS = get_config_var("OS_CLASS")
if OS_CLASS=='WIN32':
    _prefix, _suffix = '', '.dll'
elif OS_CLASS=='Darwin':
    _prefix, _suffix = 'lib', '.dylib'
else:
    _prefix, _suffix = 'lib', '.so'

include_path = os.path.join(base_path, 'include')
lib_path = os.path.join(base_path, 'lib')

# hackyness to allow deps between libraries to be found
# on targets where this can't be encoded into the library binaries.
if OS_CLASS=='WIN32':
    # windows has nothing like -rpath let alone the $ORIGIN trick
    # so we must add our /lib directory, which contains the .dlls, to %PATH%
    os.environ['PATH'] = '%s%s%s%s'%(lib_path, os.pathsep, os.environ.get('PATH', ''), os.pathsep)
    try:
        # Fix for Python 3.8 under Windows. See: https://docs.python.org/3.8/whatsnew/3.8.html#bpo-36085-whatsnew
        os.add_dll_directory(lib_path)
    except Exception:
        pass

def get_lib(name):
    """Fetch library path by name.

    :param str name: A library name.  eg. 'Com' or 'ca'.
    :returns: The full absolute path to this library.
    """
    # For ELF and MACH-O targets we are including two full copies
    # of all libraries (can't store symlinks in zip files).
    # To help maintain the sanity of those using ctypes, we always return
    # the fully qualified name, which will be the one loaded implicitly
    # by other libraries.

    base = os.path.join(lib_path, '%s%s%s'%(_prefix, name, _suffix))
    if OS_CLASS=='WIN32':
        # eg. /path/to/Com.dll
        return os.path.join(lib_path, '%s%s%s'%(_prefix, name, _suffix))
    elif OS_CLASS=='Darwin':
        # eg. /path/to/libCom.X.dylib
        return glob(os.path.join(lib_path, '%s%s.*%s'%(_prefix, name, _suffix)))[0]
    else:
        # eg. /path/to/libCom.so.X
        return glob(os.path.join(lib_path, '%s%s%s.*'%(_prefix, name, _suffix)))[0]


libraries = ('Com', 'ca', 'pvData', 'pvAccess', 'pvAccessCA', 'dbCore', 'dbRecStd')



"""I automatically configure for use with pyepics

eg. attempt to use libca from this packages if available.

    try:
        import epicscorelibs.path.pyepics
    except ImportError:
        pass
    import epics
    ...
"""

import os, sys, warnings

from . import get_lib

if 'epics.ca' in sys.modules:
    warnings.warn("epicscorelibs.path.pyepics must be imported before epics.ca to have effect")

# don't override if already set
if 'PYEPICS_LIBCA' not in os.environ:
    os.environ['PYEPICS_LIBCA'] = get_lib('ca')



'''Functions imported directly from libca.

See http://www.aps.anl.gov/epics/base/R3-14/11-docs/CAref.html for detailed
documentation of the functions below.

This module is a thin wrapper over the cadef.h file to be found in
    $EPICS_BASE/include/cadef.h
'''

__all__ = [
    # Event type notification codes for camonitor
    'DBE_VALUE',        # Notify normal value changes
    'DBE_LOG',          # Notify archival value changes
    'DBE_ALARM',        # Notify alarm state changes
    'DBE_PROPERTY',     # Notify property change events (3.14.11 and later)
]


import ctypes

from epicscorelibs import path

from . import py23

if path.OS_CLASS == "WIN32":
    # On windows, use stdcall calling convention for libca
    libca = ctypes.WinDLL(path.get_lib("ca"), ctypes.RTLD_GLOBAL)
else:
    libca = ctypes.CDLL(path.get_lib("ca"), ctypes.RTLD_GLOBAL)


# -----------------------------------------------------------------------------
#   Enumeration and error code definitions.


# Flags used to identify notification events to request for subscription.
DBE_VALUE   = 1
DBE_LOG     = 2
DBE_ALARM   = 4
DBE_PROPERTY = 8

# Connection state as passed to connection handler
CA_OP_CONN_UP   = 6
CA_OP_CONN_DOWN = 7

# Status codes as returned by virtually all ca_ routines.  We only specially
# handle normal return, timeout, or disconnection.
ECA_NORMAL = 1
ECA_TIMEOUT = 80
ECA_DISCONN = 192


# -----------------------------------------------------------------------------
#   Function wrappers.
#   Wrappers for callbacks from C back into Python.


# Event handler for ca_array_get_callback, ca_array_put_callback,
# ca_create_subscription.  The event handler is called when the action
# completes or the data is available.
class event_handler_args(ctypes.Structure):
    _fields_ = [
        ('usr',     ctypes.py_object),  # Associated private data
        ('chid',    ctypes.c_void_p),   # Channel ID for this request
        ('type',    ctypes.c_long),     # DBR type of data returned
        ('count',   ctypes.c_long),     # Number of data points returned
        ('raw_dbr', ctypes.c_void_p),   # Pointer to raw dbr array
        ('status',  ctypes.c_int)]      # ECA_ status code of operation
event_handler = ctypes.CFUNCTYPE(None, event_handler_args)


# Exception handler, called to report asynchronous errors that have no other
# report path.
class exception_handler_args(ctypes.Structure):
    _fields_ = [
        ('usr',     ctypes.c_void_p),   # Associated private data
        ('chid',    ctypes.c_void_p),   # Channel ID or NULL
        ('type',    ctypes.c_long),     # Data type requested
        ('count',   ctypes.c_long),     # Number of data points requested
        ('addr',    ctypes.c_void_p),   # User address for GET operation
        ('stat',    ctypes.c_long),     # Channel access status code
        ('op',      ctypes.c_long),     # CA_OP_ operation code
        ('ctx',     ctypes.c_char_p),   # Context information string
        ('pFile',   ctypes.c_char_p),   # Location in source: file name
        ('lineNo',  ctypes.c_uint)]     #             ... and line number
exception_handler = ctypes.CFUNCTYPE(None, exception_handler_args)


# Connection handler, used to report channel connection status.
class ca_connection_handler_args(ctypes.Structure):
    _fields_ = [
        ('chid',    ctypes.c_void_p),
        ('op',      ctypes.c_long)]
connection_handler = ctypes.CFUNCTYPE(None, ca_connection_handler_args)



# -----------------------------------------------------------------------------
#   Exception handling support
#   The following defines exceptions associated with libca errors together
# with error handling wrappers to detect the errors and raise the appropriate
# exceptions.


class Disconnected(Exception):
    '''The channel is disconnected.'''
    def __init__(self, chid):
        self.name = ca_name(chid)
    def __str__(self):
        return 'Channel %s disconnected' % self.name


class CAException(Exception):
    '''Exception in response to calling ca_ method.'''
    def __init__(self, status, function):
        self.status = status
        self.function = function
    def __str__(self):
        return '%s calling %s' % (
            ca_message(self.status), self.function.__name__)



# For routines which are simply expected to succeed this routine should be
# assigned to the routine's errcheck attribute.
def expect_ECA_NORMAL(status, function, args):
    if status != ECA_NORMAL:
        raise CAException(status, function)


# Routine for testing functions which interrogate the status of a channel and
# return a special sentinel value when the channel is disconnected.
def expect_connected(sentinel, chid = 0):
    '''sentinel is the disconnected notification value to check for, and chid
    is the position in the argument list of the channel being inspected.'''
    def expects(result, function, args):
        if result == sentinel:
            raise Disconnected(args[chid])
        else:
            return result
    return expects


# We set things up so that the data type returned by ca_puser() below is a
# Python object (to be precise, the argument passed in to ca_create_channel),
# but this means that we need to get the reference counting right.
#    The ctypes library assumes when a routine returns a py_object type that
# the called routine has already incremented the object's reference count.
# If the object is just a void* then this isn't going to be so: this routine
# when assigned to errcheck will resolve this problem.
def convert_py_object(object, function, args):
    ctypes.pythonapi.Py_IncRef(object)
    return object


# Let the library know that Py_IncRef and Py_DecRef take a raw PyObject*
# (otherwise it converts the argument to a fresh int() object which receives
# the reference count: not so useful!)
ctypes.pythonapi.Py_IncRef.argtypes = [ctypes.py_object]
ctypes.pythonapi.Py_DecRef.argtypes = [ctypes.py_object]



# -----------------------------------------------------------------------------
#   libca function definitions


#   error_message = ca_message(status_code)
#
# Converts channel access status code (an int) into a printable error message.
ca_message = libca.ca_message
ca_message.argtypes = [ctypes.c_long]
ca_message.restype = ctypes.c_char_p
ca_message.errcheck = py23.auto_decode


#   channel_name = ca_name(channel)
#
# Returns name associated with channel id
ca_name = libca.ca_name
ca_name.argtypes = [ctypes.c_void_p]
ca_name.restype = ctypes.c_char_p
ca_name.errcheck = py23.auto_decode


#   @exception_handler
#   def handler(args): ...
#
#   ca_add_exception_event(handler, context)
#
# Adds global exception handler: called for all asynchronous errors.
ca_add_exception_event = libca.ca_add_exception_event
ca_add_exception_event.argtypes = [exception_handler, ctypes.c_void_p]
ca_add_exception_event.errcheck = expect_ECA_NORMAL


#   chtype = ca_field_type(channel_id)
#
# Returns the native data type (a dbf_ code) of the data associated with the
# channel, or returns TYPENOTCONN if the channel is not connected.
TYPENOTCONN = -1
ca_field_type = libca.ca_field_type
ca_field_type.argtypes = [ctypes.c_void_p]
ca_field_type.restype = ctypes.c_short
ca_field_type.errcheck = expect_connected(TYPENOTCONN)


#   count = ca_element_count(channel_id)
#
# Returns the array element count for the data array associated with the
# channel.  Returns 0 if the channel is not connected.
ca_element_count = libca.ca_element_count
ca_element_count.argtypes = [ctypes.c_void_p]
ca_element_count.errcheck = expect_connected(0)


#   @connection_handler
#   def handler(args): ...
#
#   ca_create_channel(
#       pv, handler, context, priority, byref(channel_id))
#
# Initiates the creation of a channel with name pv.  The handler will be
# called when the state of the channel changes.  The context argument can be
# recovered by calling ca_puser(channel_id).
ca_create_channel = libca.ca_create_channel
ca_create_channel.argtypes = [
    py23.auto_encode, connection_handler, ctypes.py_object,
    ctypes.c_int, ctypes.c_void_p]
ca_create_channel.errcheck = expect_ECA_NORMAL


#   status = ca_clear_channel(channel_id)
#
# Closes the given channel.
ca_clear_channel = libca.ca_clear_channel
ca_clear_channel.argtypes = [ctypes.c_void_p]
ca_clear_channel.errcheck = expect_ECA_NORMAL


#   puser = ca_puser(channel_id)
#
# Returns the private user context associated with the channel.
ca_puser = libca.ca_puser
ca_puser.argtypes = [ctypes.c_void_p]
ca_puser.restype = ctypes.py_object
ca_puser.errcheck = convert_py_object


#   @event_handler
#   def handler(args): ...
#
#   ca_array_get_callback(
#       channel_type, count, channel_id, handler, context)
#
# Makes a request to receive data from the given channel.  The handler will
# be called if data is retrieved or the channel becomes disconnected.
ca_array_get_callback = libca.ca_array_get_callback
ca_array_get_callback.argtypes = [
    ctypes.c_long, ctypes.c_long, ctypes.c_void_p, event_handler,
    ctypes.py_object]
ca_array_get_callback.errcheck = expect_ECA_NORMAL


#   @event_handler
#   def handler(args): ...
#
#   ca_array_put_callback(
#       channel_type, count, channel_id, value, handler, context)
#
# Writes data to the given channel.  The handler is called once server side
# processing is complete, or if a failure occurs.
ca_array_put_callback = libca.ca_array_put_callback
ca_array_put_callback.argtypes = [
    ctypes.c_long, ctypes.c_long, ctypes.c_void_p, ctypes.c_void_p,
    event_handler, ctypes.py_object]
ca_array_put_callback.errcheck = expect_ECA_NORMAL


#   ca_array_put(channel_type, count, channel_id, value)
#
# Writes data to the given channel, returning immediately.  There may be no
# notification if this fails.
ca_array_put = libca.ca_array_put
ca_array_put.argtypes = [
    ctypes.c_long, ctypes.c_long, ctypes.c_void_p, ctypes.c_void_p]
ca_array_put.errcheck = expect_ECA_NORMAL


#   @event_handler
#   def handler(args): ...
#
#   ca_create_subscription(
#       channel_type, count, channel_id, event_mask, handler, context,
#       byref(event_id))
#
# Creates a subscription to receive notification whenever significant changes
# occur to a channel.  Arguments are:
#
#   channel_type: DBR type of data to be passed to handler callback
#
#   count: the number of points to be read from the underlying data array in
#       the channel, or 0 to specify native length.
#
#   channel_id: the identifier of a previously created channel.  Note that
#       subscriptions can be created on a disconnected channel.
#
#   event_mask: mask of events to receive, any combination of DBE_ values.
#
#   handler: callback function to receive updates, receives an
#       event_handler_args structure as argument.
#
#   context: callback context passed as .usr field of handler arguments.
#
#   event_id: an identifier for this subscription, used for subsequently
#       cancelling this subscription.
ca_create_subscription = libca.ca_create_subscription
ca_create_subscription.argtypes = [
    ctypes.c_long, ctypes.c_long, ctypes.c_void_p, ctypes.c_long,
    event_handler, ctypes.py_object, ctypes.c_void_p]
ca_create_subscription.errcheck = expect_ECA_NORMAL


#   ca_clear_subscription(event_id)
#
# Cancels a previously established subscription using the returned event_id.
ca_clear_subscription = libca.ca_clear_subscription
ca_clear_subscription.argtypes = [ctypes.c_void_p]
ca_clear_subscription.errcheck = expect_ECA_NORMAL


#   ca_context_create(enable_preemptive_callback)
#
# To be called on initialisation, specifying whether asynchronous callbacks
# are to be enabled.
ca_context_create = libca.ca_context_create
ca_context_create.argtypes = [ctypes.c_int]
ca_context_create.errcheck = expect_ECA_NORMAL


#   ca_context_destroy()
#
# To be called at exit.
ca_context_destroy = libca.ca_context_destroy
ca_context_destroy.argtypes = []


#   status = ca_pend_event(timeout)
#
# Flushes the send buffer and processes background activities until the
# specified timeout (in seconds) expires.
ca_pend_event = libca.ca_pend_event
ca_pend_event.argtypes = [ctypes.c_double]


#   ca_flush_io()
#
# Flush outstanding IO requests to the server.  Needs to be called after CA
# calls which require interaction with a CA server.
ca_flush_io = libca.ca_flush_io
ca_flush_io.argtypes = []
ca_flush_io.errcheck = expect_ECA_NORMAL


#   state = ca_state(channel_id)
#
# Returns an enumeration indicating the state of the channel.
cs_never_conn = 0
cs_prev_conn = 1
cs_conn = 2
cs_closed = 3
ca_state = libca.ca_state
ca_state.argtypes = [ctypes.c_void_p]


#   host = ca_host_name(channel_id)
#
# Returns the host name of the connected server
ca_host_name = libca.ca_host_name
ca_host_name.argtypes = [ctypes.c_void_p]
ca_host_name.restype = ctypes.c_char_p
ca_host_name.errcheck = py23.auto_decode


#   read = ca_read_access(channel_id)
#   write = ca_write_access(channel_id)
#
# Returns whether the channel can be read or written by this client.
ca_read_access = libca.ca_read_access
ca_read_access.argtypes = [ctypes.c_void_p]
ca_read_access.restype = bool

ca_write_access = libca.ca_write_access
ca_write_access.argtypes = [ctypes.c_void_p]
ca_write_access.restype = bool



"""This package holds the non-python libraries which consitute EPICS Base.

https://epics.anl.gov/

See epicscorelibs.license for terms
"""

license = """
Copyright (c) 1991-2011 UChicago Argonne LLC.
Copyright (c) 1991-2006 The Regents of the University of California.
Copyright (c) 2006-2011. Los Alamos National Security, LLC. Some of this
material was produced under U.S. Government contract DE-AC52-06NA25396
for Los Alamos National Laboratory (LANL), which is operated by Los Alamos
National Security, LLC for the U.S. Department of Energy.

EPICS BASE is distributed subject to the following license conditions:

 SOFTWARE LICENSE AGREEMENT
 Software: EPICS BASE

 1. The "Software", below, refers to EPICS BASE (in either source code, or
    binary form and accompanying documentation). Each licensee is
    addressed as "you" or "Licensee."

 2. The copyright holders shown above and their third-party licensors
    hereby grant Licensee a royalty-free nonexclusive license, subject to
    the limitations stated herein and U.S. Government license rights.

 3. You may modify and make a copy or copies of the Software for use
    within your organization, if you meet the following conditions:
      a. Copies in source code must include the copyright notice and this
         Software License Agreement.
      b. Copies in binary form must include the copyright notice and this
         Software License Agreement in the documentation and/or other
         materials provided with the copy.

 4. You may modify a copy or copies of the Software or any portion of it,
    thus forming a work based on the Software, and distribute copies of
    such work outside your organization, if you meet all of the following
    conditions:
      a. Copies in source code must include the copyright notice and this
         Software License Agreement;
      b. Copies in binary form must include the copyright notice and this
         Software License Agreement in the documentation and/or other
         materials provided with the copy;
      c. Modified copies and works based on the Software must carry
         prominent notices stating that you changed specified portions of
         the Software.

 5. Portions of the Software resulted from work developed under a U.S.
    Government contract and are subject to the following license: the
    Government is granted for itself and others acting on its behalf a
    paid-up, nonexclusive, irrevocable worldwide license in this computer
    software to reproduce, prepare derivative works, and perform publicly
    and display publicly.

 6. WARRANTY DISCLAIMER. THE SOFTWARE IS SUPPLIED "AS IS" WITHOUT WARRANTY
    OF ANY KIND. THE COPYRIGHT HOLDERS, THEIR THIRD PARTY LICENSORS, THE
    UNITED STATES, THE UNITED STATES DEPARTMENT OF ENERGY, AND THEIR
    EMPLOYEES: (1) DISCLAIM ANY WARRANTIES, EXPRESS OR IMPLIED, INCLUDING
    BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
    FOR A PARTICULAR PURPOSE, TITLE OR NON-INFRINGEMENT, (2) DO NOT ASSUME
    ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS,
    OR USEFULNESS OF THE SOFTWARE, (3) DO NOT REPRESENT THAT USE OF THE
    SOFTWARE WOULD NOT INFRINGE PRIVATELY OWNED RIGHTS, (4) DO NOT WARRANT
    THAT THE SOFTWARE WILL FUNCTION UNINTERRUPTED, THAT IT IS ERROR-FREE
    OR THAT ANY ERRORS WILL BE CORRECTED.

 7. LIMITATION OF LIABILITY. IN NO EVENT WILL THE COPYRIGHT HOLDERS, THEIR
    THIRD PARTY LICENSORS, THE UNITED STATES, THE UNITED STATES DEPARTMENT
    OF ENERGY, OR THEIR EMPLOYEES: BE LIABLE FOR ANY INDIRECT, INCIDENTAL,
    CONSEQUENTIAL, SPECIAL OR PUNITIVE DAMAGES OF ANY KIND OR NATURE,
    INCLUDING BUT NOT LIMITED TO LOSS OF PROFITS OR LOSS OF DATA, FOR ANY
    REASON WHATSOEVER, WHETHER SUCH LIABILITY IS ASSERTED ON THE BASIS OF
    CONTRACT, TORT (INCLUDING NEGLIGENCE OR STRICT LIABILITY), OR
    OTHERWISE, EVEN IF ANY OF SAID PARTIES HAS BEEN WARNED OF THE
    POSSIBILITY OF SUCH LOSS OR DAMAGES.

________________________________________________________________________

This software is in part copyrighted by the BERLINER SPEICHERRING
GESELLSCHAFT FUER SYNCHROTRONSTRAHLUNG M.B.H. (BESSY), BERLIN, GERMANY.

In no event shall BESSY be liable to any party for direct, indirect,
special, incidental, or consequential damages arising out of the use of
this software, its documentation, or any derivatives thereof, even if
BESSY has been advised of the possibility of such damage.

BESSY specifically disclaims any warranties, including, but not limited
to, the implied warranties of merchantability, fitness for a particular
purpose, and non-infringement.  This software is provided on an "as is"
basis, and BESSY has no obligation to provide maintenance, support,
updates, enhancements, or modifications.
________________________________________________________________________
"""



# Some simple python2/python3 compatibility tricks.
#
# Some of this is drastically reduced from six.py

import sys
import ctypes


# Exception handling
if sys.version_info < (3,):
    def raise_from(exception, source):
        raise exception

    exec('''
def raise_with_traceback(result):
    raise result[0], result[1], result[2]
''')

else:
    exec('''
def raise_from(exception, source):
    try:
        raise exception from source
    finally:
        exception = None
''')

    def raise_with_traceback(result):
        raise result[1].with_traceback(result[2])


# c_char_p conversion
if sys.version_info < (3,):
    auto_encode = ctypes.c_char_p
    def auto_decode(result, func, args):
        return result
    def decode(s):
        return s

else:
    class auto_encode(ctypes.c_char_p):
        @classmethod
        def from_param(cls, value):
            if value is None:
                return value
            else:
                return value.encode('UTF-8')

    def auto_decode(result, func, args):
        if result is None:
            return result
        else:
            return result.decode('UTF-8', 'replace')

    def decode(s):
        return s.decode('UTF-8', 'replace')



#!/usr/bin/env python

from __future__ import print_function

import os, sys, platform, re
from glob import glob
from collections import defaultdict

from setuptools import Command, Distribution
from setuptools.command import install
from distutils.command import build

from setuptools_dso import Extension, setup, DSO, build_dso

from distutils import log


mydir = os.path.abspath(os.path.dirname(__file__))

# I will use some of what I'm about to install
sys.path.insert(0, os.path.join(mydir, 'src', 'python'))

from epicscorelibs.config import get_config_var


# our choice of version suffix is constrained by PEP 440
# so we always append .99.ABI.SRC to most recent upstream version
# the following line is matched from cibuild.py
package_version = '7.0.5.99.1.0a2'

assert package_version.split('.')[-3]=='99', package_version

abi_version = '.'.join(package_version.split('.')[:-1]) # strip off the last component

OS_CLASS = get_config_var('OS_CLASS')

# Extract error symbol definitoins from C headers
# eg.
# | #define M_asLib                   (501 << 16)  /* Access Security */
# | #define S_asLib_asNotActive       (M_asLib|10) /*access security is not active*/
#            <macro>                  <value>        <desc>
err_def = re.compile(r'''
    ^\s* \# \s* define \s+ (?P<macro>[SM]_[A-Za-z0-9_]+)
        \s+ \( (?P<value>[^)]*) \)
        \s* /\* (?P<desc>.*) \*/ \s*$
''', re.X)

def readmake(defs, *parts):
    """Read "makefile" containing only definitions (VAR = value)
    """
    with open(os.path.join(mydir, *parts), 'r') as F:
        for line in F:
            line = line.strip()
            if line[:1] in ('', '#'):
                continue

            key, sep, value = line.partition('=')
            if sep!='=':
                continue
            defs[key.strip()] = value.strip()

class GenVersionError(Command):
    """Generate epicsVersion.h, config parameter, and error symbol tables
    """
    user_options = [
        ('build-temp=', 't',
         "directory for temporary files (build by-products)"),
    ]

    def initialize_options(self):
        self.build_temp = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('build_temp', 'build_temp'),
                                  )

    def run(self):
        log.info("In GetVersionError")
        self.mkpath(self.build_temp)
        defs = {}
        readmake(defs, 'configure', 'CONFIG_BASE_VERSION')
        defs['EPICS_SITE_VERSION'] = 'pipi'

        outfile = os.path.join(self.build_temp, 'epicsVersion.h')
        if self.dry_run:
            log.info("Would write %s", outfile)
        else:
            log.info("Writing %s", outfile)
            with open(outfile, 'w') as F:
                F.write("""
#ifndef INC_epicsVersion_H
#define INC_epicsVersion_H

#define EPICS_VERSION        %(EPICS_VERSION)s
#define EPICS_REVISION       %(EPICS_REVISION)s
#define EPICS_MODIFICATION   %(EPICS_MODIFICATION)s
#define EPICS_PATCH_LEVEL    %(EPICS_PATCH_LEVEL)s
#define EPICS_DEV_SNAPSHOT   "%(EPICS_DEV_SNAPSHOT)s"
#define EPICS_SITE_VERSION   "%(EPICS_SITE_VERSION)s"

#define EPICS_VERSION_SHORT "%(EPICS_VERSION)s.%(EPICS_REVISION)s.%(EPICS_MODIFICATION)s.%(EPICS_PATCH_LEVEL)s"
#define EPICS_VERSION_FULL "%(EPICS_VERSION)s.%(EPICS_REVISION)s.%(EPICS_MODIFICATION)s.%(EPICS_PATCH_LEVEL)s%(EPICS_DEV_SNAPSHOT)s"
#define EPICS_VERSION_STRING "EPICS %(EPICS_VERSION)s.%(EPICS_REVISION)s.%(EPICS_MODIFICATION)s.%(EPICS_PATCH_LEVEL)s%(EPICS_DEV_SNAPSHOT)s"
#define epicsReleaseVersion  "EPICS %(EPICS_VERSION)s.%(EPICS_REVISION)s.%(EPICS_MODIFICATION)s.%(EPICS_PATCH_LEVEL)s%(EPICS_DEV_SNAPSHOT)s"

#ifndef VERSION_INT
#  define VERSION_INT(V,R,M,P) ( ((V)<<24) | ((R)<<16) | ((M)<<8) | (P))
#endif
#define EPICS_VERSION_INT VERSION_INT(EPICS_VERSION, EPICS_REVISION, EPICS_MODIFICATION, EPICS_PATCH_LEVEL)

#endif /* INC_epicsVersion_H */
"""%defs)

        defs = {}
        defs['EPICS_BUILD_TARGET_ARCH'] = get_config_var('EPICS_HOST_ARCH')
        readmake(defs, 'configure', 'CONFIG_ENV')
        readmake(defs, 'configure', 'CONFIG_SITE_ENV')

        outfile = os.path.join(self.build_temp, 'generated.c')
        if self.dry_run:
            log.info("Would write %s", outfile)
        else:
            log.info("Writing %s", outfile)
            with open(outfile, 'w') as F:
                F.write("""
#include <stddef.h>
#include "envDefs.h"
""")

                for N, V in defs.items():
                    F.write('const ENV_PARAM %s = {"%s", "%s"};\n'%(N,N,V.strip('"')))

                F.write('const ENV_PARAM* env_param_list[] = {\n')
                for N, V in defs.items():
                    F.write('    &%s,\n'%N)
                F.write('    NULL\n};\n')

                F.write("""
#include "errMdef.h"
#include "errSymTbl.h"
#include "dbDefs.h"
""")

                defM, defS = [], []
                for errfile in self.distribution.x_errtable or []:
                    with open(os.path.normcase(errfile), 'r') as I:
                        for line in I:
                            M = err_def.match(line)
                            if M is None:
                                continue
                            D = M.groupdict()
                            # TODO: escape 'desc'?
                            if D['macro'][0]=='M':
                                defM.append(D)
                            elif D['macro'][0]=='S':
                                defS.append(D)

                if len(defM)==0 or len(defS)==0:
                    log.warn("Warning: generated error string table is empty!")
                    # MSVC doesn't like zero length C arrays
                    # so give it something
                    defS.append({'macro':'S_generic', 'value':'1', 'desc':'placeholder'})

                for D in defM:
                    F.write("#ifndef %(macro)s\n#  define %(macro)s (%(value)s) /*%(desc)s*/\n#endif\n"%D)

                for D in defS:
                    F.write("#define %(macro)s (%(value)s) /*%(desc)s*/\n"%D)

                F.write("""
static ERRSYMBOL symbols[] = {
""")

                for D in defS:
                    F.write("    { (long)%(macro)s, \"%(desc)s\" },\n"%D)

                F.write("""
}; /* TODO */
static ERRSYMTAB symTbl = {NELEMENTS(symbols), symbols};
ERRSYMTAB_ID errSymTbl = &symTbl;
""")

        outfile = os.path.join(self.build_temp, 'epicsVCS.h')
        if self.dry_run:
            log.info("Would write %s", outfile)
        else:
            log.info("Writing %s", outfile)
            with open(outfile, 'w') as F:
                F.write("""
#ifndef EPICS_VCS_VERSION
  #define EPICS_VCS_VERSION "%s"
#endif
"""%package_version)
        

Distribution.x_errtable = None

class Expand(Command):
    user_options = [
        ('build-temp=', 't',
         "directory for temporary files (build by-products)"),
    ]

    def initialize_options(self):
        self.build_temp = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('build_temp', 'build_temp'),
                                  )

    def run(self):
        log.info("In Expand")
        for template, outname, files in self.distribution.x_expand or []:
            template = os.path.normcase(template)

            outname = os.path.join(self.build_temp, os.path.normcase(outname))
            log.info("expand %s -> %s", template, outname)

            with open(template, 'r') as F:
                tmpl = F.read()

            defs = {}
            for fname in files:
                readmake(defs, fname)

            def repl(M):
                return defs[M.group(1)]

            out = re.sub(r'@([^@]*)@', repl, tmpl)

            if not self.dry_run:
                self.mkpath(os.path.dirname(outname))
                with open(outname, 'w') as F:
                    F.write(out)

Distribution.x_expand = None

class MakeAPIH(Command):
    user_options = [
        ('build-lib=', 't',
         "directory for temporary files (build by-products)"),
    ]

    def initialize_options(self):
        self.build_lib = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('build_lib', 'build_lib'),
                                  )

    def run(self):
        for stem, outname in self.distribution.x_api_h or []:
            outname = os.path.join(self.build_lib, 'epicscorelibs', 'include', os.path.basename(outname))
            log.info('make %s -> %s', stem, outname)
            out = '''
#ifndef %(guard)s
#define %(guard)s

#if defined(_WIN32) || defined(__CYGWIN__)

#  if !defined(epicsStdCall)
#    define epicsStdCall __stdcall
#  endif

#  if defined(BUILDING_%(stem)s_API) && defined(EPICS_BUILD_DLL)
/* Building library as dll */
#    define %(STEM)s_API __declspec(dllexport)
#  elif !defined(BUILDING_%(stem)s_API) && defined(EPICS_CALL_DLL)
/* Calling library in dll form */
#    define %(STEM)s_API __declspec(dllimport)
#  endif

#elif __GNUC__ >= 4
#  define %(STEM)s_API __attribute__ ((visibility("default")))
#endif

#if !defined(%(STEM)s_API)
#  define %(STEM)s_API
#endif

#if !defined(epicsStdCall)
#  define epicsStdCall
#endif

#endif /* %(guard)s */
'''%{'stem':stem, 'STEM':stem.upper(), 'guard':'INC_'+stem+'API_H'}

            if not self.dry_run:
                self.mkpath(os.path.dirname(outname))
                with open(outname, 'w') as F:
                    F.write(out)

Distribution.x_api_h = None

class InstallHeaders(Command):
    user_options = [
        ('build-lib=', 't',
         "directory for temporary files (build by-products)"),
        ('build-temp=', 't',
         "directory for temporary files (build by-products)"),
    ]
    def initialize_options(self):
        self.build_lib = None
        self.build_temp = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('build_lib', 'build_lib'),
                                   ('build_temp', 'build_temp'),
                                  )
    def run(self):
        log.info("In InstallHeaders")
        for pkg, files in (self.distribution.x_headers or {}).items():
            print("Install headers for", pkg)
            pkgdir = os.path.join(*[self.build_lib]+pkg.split('.'))
            self.mkpath(pkgdir)

            for fname in files:
                if isinstance(fname, str):
                    fout = fname
                else:
                    fout, fname = fname
                if not os.path.isfile(fname):
                    fname = os.path.join(self.build_temp, fname)

                self.copy_file(fname,
                               os.path.join(pkgdir, os.path.basename(fout)))

Distribution.x_headers = None

build_dso.sub_commands.extend([
    ('build_expand', lambda self:True),
    ('build_apih', lambda self:True),
    ('build_generated', lambda self:True),
    ('install_epics_headers', lambda self:True),
])

def readlists(name, prefix):
    prefix = os.path.normcase(prefix)
    groups = ['all', get_config_var('OS_CLASS')]

    ret = []

    for part in ('source', 'include', 'header'):
        R = []

        for grp in groups:
            fname = os.path.join(mydir, 'setup.py.d', '.'.join([name, grp, part]))
            if os.path.isfile(fname):
                with open(fname, 'r') as F:
                    for line in F:
                        line = line.strip()
                        if line[:1] in ('', '#'):
                            continue
                        R.append(os.path.normcase(line))

        R = [os.path.join(prefix, F) for F in R]
        ret.append(R)

    return tuple(ret)

modules = []
headers = ['epicsVersion.h']

local_defs = [('USE_TYPED_RSET', None)]

def build_module(name, srcdir, defs=[], deps=[], srcs=[], soversion=None):
    #print("Include EPICS module %s in %s"%(name, srcdir))
    src, inc, hdr = readlists(name, srcdir)
    if len(src)==0:
        raise RuntimeError("module %s has no source"%name)
    defs = get_config_var('CPPFLAGS') + local_defs + defs

    MOD = DSO(
        name='epicscorelibs.lib.'+name,
        sources = srcs+src,
        include_dirs = [os.path.join('epicscorelibs', 'include')] + inc + ['.'],
        define_macros = defs,
        dsos = ['epicscorelibs.lib.'+D for D in deps],
        libraries = get_config_var('LDADD'),
        extra_link_args = get_config_var('LDFLAGS'),
        lang_compile_args = {
            'c':get_config_var('CFLAGS'),
            'c++':get_config_var('CXXFLAGS'),
        },
        soversion=soversion or abi_version,
    )

    modules.append(MOD)
    headers.extend(hdr)

build_module('Com', 'modules/libcom/src',
             defs=[('BUILDING_libCom_API', None), ('EPICS_COMMANDLINE_LIBRARY', 'EPICS_COMMANDLINE_LIBRARY_EPICS')],
             srcs=['generated.c'],
)
build_module('ca', 'modules/ca/src',
             defs=[('BUILDING_libCa_API', None)],
             deps=['Com'],
)
build_module('dbCore', 'modules/database/src/ioc',
             defs=[('BUILDING_dbCore_API', None)],
             deps=['ca','Com'],
)
build_module('dbRecStd', 'modules/database/src/std',
             defs=[('BUILDING_dbRecStd_API', None)],
             deps=['dbCore', 'ca','Com'],
)
build_module('pvData', 'modules/pvData',
             deps=['ca', 'Com'],
)
build_module('pvAccess', 'modules/pvAccess',
             deps=['pvData', 'Com'],
)
build_module('pvAccessCA', 'modules/pvAccess',
             deps=['pvAccess', 'pvData', 'ca', 'Com'],
)
build_module('pvAccessIOC', 'modules/pvAccess',
             deps=['pvAccess', 'pvData', 'dbRecStd', 'dbCore', 'ca', 'Com'],
)
build_module('qsrv', 'modules/pva2pva',
             defs=[('QSRV_API_BUILDING',None)],
             deps=['pvAccess', 'pvData', 'dbRecStd', 'dbCore', 'ca', 'Com'],
)

assert len(headers)>0

_header_dirs = set(['pv', 'pva', 'valgrind'])

def proc_headers(headers):
    """Take list of headers sort by installed sub-directory

    {'epicscorelibs.include.PATH': ['path/of/header.h', ...], ...}
    """
    groups = defaultdict(list)
    for header in headers:
        hdir, hname = os.path.split(header)
        hmod = ['epicscorelibs', 'include']
        hdir, candidate = os.path.split(hdir)
        if candidate in _header_dirs:
            hmod.append(candidate)
        groups['.'.join(hmod)].append(header)

    return groups

headers = proc_headers(headers)

headers['epicscorelibs.dbd'] = [] \
    + glob('modules/database/src/ioc/*/*.dbd') \
    + glob('modules/database/src/std/*/*.dbd') \
    + glob('modules/database/src/ioc/generated/*.dbd') \
    + glob('modules/database/src/std/generated/*.dbd') \
    + glob('modules/pvAccess/src/ioc/*.dbd') \
    + [('qsrv.dbd', 'modules/pva2pva/pdbApp/qsrv-new.dbd')]

# a dummy extension so that bdist_wheel will recognise this package
# as containing binaries.
ext = Extension(
    name='epicscorelibs._base',
    sources = ['src/python/epicscorelibs/base.cpp'],
    define_macros = get_config_var('CPPFLAGS'),
    extra_compile_args = get_config_var('CXXFLAGS'),
    extra_link_args = get_config_var('LDFLAGS'),
)

setup(
    name='epicscorelibs',
    version=package_version,
    description="The EPICS Core libraries for use by python modules",
    long_description="""The EPICS (Experimental Physics and Industrial Control System) core libraries
for use by python modules.  Either dynamically with ctypes or statically by compiled extension.
""",
    url='https://github.com/mdavidsaver/epicscorelibs',
    author='Michael Davidsaver',
    author_email='mdavidsaver@gmail.com',
    license='EPICS',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'License :: Freely Distributable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Distributed Computing',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords='epics scada',
    project_urls = {
        'Source': 'https://github.com/mdavidsaver/epicscorelibs',
        'Tracker': 'https://github.com/mdavidsaver/epicscorelibs/issues',
        'Upstream': 'https://epics.anl.gov/',
    },
    python_requires='>=2.7',
    install_requires=[
        'setuptools', # needed at runtime for 'pkg_resources'
        'numpy', # needed for epicscorelibs.ca.dbr
    ],

    packages=['epicscorelibs', 'epicscorelibs.path', 'epicscorelibs.test', 'epicscorelibs.ca'],
    package_dir={'':os.path.join('src','python')},
    package_data={
        '':['*.pxd'],
        'dbd':[
            'modules/database/src/ioc/*/*.dbd',
            'modules/database/src/std/*/*.dbd',
            'modules/database/src/ioc/generated/*.dbd',
            'modules/database/src/std/generated/*.dbd',
        ],
    },
    ext_modules = [ext],
    x_dsos = modules,
    x_headers = headers,
    x_api_h = [
        ('libCom', 'modules/libcom/src/libComAPI.h'),
        ('libCa', 'modules/ca/src/client/libCaAPI.h'),
        ('dbCore', 'modules/database/src/ioc/dbCoreAPI.h'),
        ('dbRecStd', 'modules/database/src/std/dbRecStdAPI.h'),
    ],
    x_expand = [
        ('modules/libcom/src/libComVersion.h@', 'modules/libcom/src/libComVersion.h',
         ['configure/CONFIG_LIBCOM_VERSION']),
        ('modules/database/src/ioc/databaseVersion.h@', 'modules/database/src/ioc/databaseVersion.h',
         ['configure/CONFIG_DATABASE_VERSION']),
        ('modules/pvData/src/pv/pvdVersionNum.h@', 'modules/pvData/src/pv/pvdVersionNum.h',
         ['modules/pvData/configure/CONFIG_PVDATA_VERSION']),
        ('modules/pvAccess/src/pva/pvaVersionNum.h@', 'modules/pvAccess/src/pva/pv/pvaVersionNum.h',
         ['modules/pvAccess/configure/CONFIG_PVACCESS_VERSION']),
        ('modules/pva2pva/pdbApp/pv/qsrvVersionNum.h@', 'modules/pva2pva/pdbApp/pv/qsrvVersionNum.h',
         ['modules/pva2pva/configure/CONFIG_QSRV_VERSION']),
    ],
    x_errtable = [
        "modules/libcom/src/osi/devLib.h",
        "modules/libcom/src/osi/epicsTime.h",
        #"modules/libcom/src/as/asLib.h",
        "modules/libcom/src/misc/epicsStdlib.h",
        "modules/libcom/src/pool/epicsThreadPool.h",
        "modules/libcom/src/error/errMdef.h",
    ],
    cmdclass = {
        'build_generated': GenVersionError,
        'build_expand': Expand,
        'build_apih': MakeAPIH,
        'install_epics_headers':InstallHeaders,
    },
    zip_safe = False,
)



'''Channel access datatype support.  Derived from definitions in the EPICS
header file db_access.h
'''

import sys
import ctypes
import numpy
import datetime

from . import cadef
from . import py23


__all__ = [
    # Basic DBR request codes: any one of these can be used as part of a
    # datatype request.
    'DBR_STRING',       # 40 character strings
    'DBR_SHORT',        # 16 bit signed
    'DBR_FLOAT',        # 32 bit float
    'DBR_ENUM',         # 16 bit unsigned
    'DBR_CHAR',         # 8 bit unsigned
    'DBR_LONG',         # 32 bit signed
    'DBR_DOUBLE',       # 64 bit float

    'DBR_CHAR_STR',     # Long strings as char arrays
    'DBR_CHAR_UNICODE', # Long unicode strings as char arrays
    'DBR_ENUM_STR',     # Enums as strings, default otherwise
    'DBR_CHAR_BYTES',   # Long byte strings as char arrays

    'DBR_PUT_ACKT',     # Configure global alarm acknowledgement
    'DBR_PUT_ACKS',     # Acknowledge global alarm
    'DBR_STSACK_STRING', # Returns status ack structure
    'DBR_CLASS_NAME',   # Returns record type (same as .RTYP?)

    # Data type format requests
    'FORMAT_RAW',       # Request the underlying data only
    'FORMAT_TIME',      # Request alarm status and timestamp
    'FORMAT_CTRL',      # Request graphic and control fields

    'ca_extra_fields',  # List of all possible augmented field names
]

# List of all the field names that can be added to an augmented field.
ca_extra_fields = [
    # Fields common to all data types
    'name',         # Name of the PV used to create this value
    'ok',           # True for normal data, False for error code
    'datatype',     # Underlying DBR_ code
    'element_count', # Underlying length of original data
    # Fields common to time and ctrl types
    'severity',     # Alarm severity
    'status',       # CA status code: reason for severity
    # Timestamp specific fields
    'raw_stamp',    # Unformatted timestamp in separate seconds and nsecs
    'timestamp',    # Timestamp in seconds
    'datetime',     # Timestamp converted to datetime
    # Control specific fields
    'units',        # Units for display
    'upper_disp_limit',
    'lower_disp_limit',
    'upper_alarm_limit',
    'lower_alarm_limit',
    'upper_warning_limit',
    'lower_warning_limit',
    'upper_ctrl_limit',
    'lower_ctrl_limit',
    'precision',    # Display precision for floating point values
    # Other
    'enums',        # Enumeration strings for ENUM type
]


# Standard hard-wired EPICS array sizes.
MAX_STRING_SIZE = 40        # Size of string type
MAX_UNITS_SIZE = 8          # Size of units string
MAX_ENUM_STRING_SIZE = 26   # Size of individual enumeration strings
MAX_ENUM_STATES = 16        # Numer of possible enumeration strings


ca_doc_string = \
'''All values returned from channel access are returned as "augmented"
types with extra fields.  The following fields are always present:
    name
    ok

The followiong fields are present if ok is True:
    datatype
    element_count

Depending on the request type, the following extra fields may be present:

If timestamps requested:
    status, severity,
    timestamp, raw_stamp

    The timestamp is returned in two forms:
        timestamp
            This is the time stamp in the system epoch in seconds represented
            as a double.  Rounding leads to errors at the resolution of
            sub-microseconds, so this result is rounded to the nearest
            microsecond.
        raw_stamp
            This is a tuple of the time stamp as (secs, nsec) with separate
            integer fields .secs and .nsec for the seconds and nanoseconds.

If control values requested (and datatype is not DBR_ENUM):
    status, severity, units,
    upper_disp_limit, lower_disp_limit,
    upper_alarm_limit, lower_alarm_limit,
    upper_warning_limit, lower_warning_limit,
    upper_ctrl_limit, lower_ctrl_limit,
    precision (if floating point type)

If control values requested and datatype is DBR_ENUM:
    status, severity,
    enums (list of possible enumeration strings)
'''

@property
def timestamp_to_datetime(self):
    return datetime.datetime.fromtimestamp(self.timestamp)

# Augmented array used for all return values with more than one element.
class ca_array(numpy.ndarray):
    __doc__ = ca_doc_string
    datetime = timestamp_to_datetime
    def __pos__(self):
        return numpy.array(self)
    __hash__ = None

# Augmented basic Python types used for scalar values.
class ca_str(str):
    __doc__ = ca_doc_string
    datetime = timestamp_to_datetime
    def __pos__(self):
        return str(self)


# Overlapping handling for python 2 and python 3.  We have three types with two
# different semantics: str, bytes, unicode.  In python2 str is bytes, while in
# python3 str is unicode.  We walk a delicate balancing act to get the right
# behaviour in both environments!
if sys.version_info < (3,):
    ca_bytes = ca_str
    class ca_unicode(bytes):
        __doc__ = ca_doc_string
        datetime = timestamp_to_datetime
        def __pos__(self):
            return unicode(self)
    str_char_code = 'S'
else:
    class ca_bytes(bytes):
        __doc__ = ca_doc_string
        datetime = timestamp_to_datetime
        def __pos__(self):
            return bytes(self)
    ca_unicode = ca_str
    str_char_code = 'U'
    unicode = str


class ca_int(int):
    __doc__ = ca_doc_string
    datetime = timestamp_to_datetime

class ca_float(float):
    __doc__ = ca_doc_string
    datetime = timestamp_to_datetime


# The EPICS epoch begins midnight first thing on 1st January 1990 and is in UTC.
# We convert all EPICS timestamps to the Python epoch.  This is not defined in
# the language documentation but in practice is the ANSI epoch, midnight 1st
# January 1970.  Strictly we should compute
#   EPICS_epoch = calendar.timegm((1990, 1, 1, 0, 0, 0, 0, 0, 0))
# but that pulls in an extra module dependency and the number is constant:
EPICS_epoch = 631152000             # Seconds from 1970 to 1990


class ca_timestamp(ctypes.Structure):
    _fields_ = [
        ('secs',    ctypes.c_uint32),
        ('nsec',    ctypes.c_uint32)]


# ----------------------------------------------------------------------------
#   DBR type definitions

# All the following types are used to overlay dbr data returned from channel
# access or passed into channel access.

def copy_attributes_none(self, other):
    pass

def copy_attributes_time(self, other):
    other.status = self.status
    other.severity = self.severity

    # Handling the timestamp is a little awkward.  We provide both a
    # raw_stamp and a timestamp value as there is loss of ns precision in
    # the timestamp value (represented as a double) and the raw_stamp value
    # is awkward for computation.
    secs = self.raw_stamp.secs + EPICS_epoch
    nsec = self.raw_stamp.nsec
    other.raw_stamp = (secs, nsec)
    # The timestamp is rounded to microseconds, both to avoid confusion
    # (because the ns part is rounded already) and to avoid an excruciating
    # bug in the .fromtimestamp() function.
    other.timestamp = round(secs + nsec * 1e-9, 6)

def copy_attributes_ctrl(self, other):
    other.status = self.status
    other.severity = self.severity

    other.units = py23.decode(ctypes.string_at(self.units))
    other.upper_disp_limit = self.upper_disp_limit
    other.lower_disp_limit = self.lower_disp_limit
    other.upper_alarm_limit = self.upper_alarm_limit
    other.lower_alarm_limit = self.lower_alarm_limit
    other.upper_warning_limit = self.upper_warning_limit
    other.lower_warning_limit = self.lower_warning_limit
    other.upper_ctrl_limit = self.upper_ctrl_limit
    other.lower_ctrl_limit = self.lower_ctrl_limit

    if hasattr(self, 'precision'):
        other.precision = self.precision

# This particular dtype is used for strings, and indeed identity to this
# value is used to recognise the string type!
str_dtype = numpy.dtype('S%d' % MAX_STRING_SIZE)


# Base DBR types
class dbr_string(ctypes.Structure):
    dtype = str_dtype
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', (ctypes.c_byte * MAX_STRING_SIZE) * 1)]

class dbr_short(ctypes.Structure):
    dtype = numpy.int16
    scalar = ca_int
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', ctypes.c_int16 * 1)]

class dbr_float(ctypes.Structure):
    dtype = numpy.float32
    scalar = ca_float
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', ctypes.c_float * 1)]

class dbr_enum(ctypes.Structure):
    dtype = numpy.uint16
    scalar = ca_int
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', ctypes.c_uint16 * 1)]

class dbr_char(ctypes.Structure):
    dtype = numpy.uint8
    scalar = ca_int
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', ctypes.c_uint8 * 1)]

class dbr_long(ctypes.Structure):
    dtype = numpy.int32
    scalar = ca_int
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', ctypes.c_int32 * 1)]

class dbr_double(ctypes.Structure):
    dtype = numpy.float64
    scalar = ca_float
    copy_attributes = copy_attributes_none
    _fields_ = [('raw_value', ctypes.c_double * 1)]

# DBR types with timestamps.

class dbr_time_string(ctypes.Structure):
    dtype = str_dtype
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('raw_value', (ctypes.c_byte * MAX_STRING_SIZE) * 1)]

class dbr_time_short(ctypes.Structure):
    dtype = numpy.int16
    scalar = ca_int
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('RISC_pad',  ctypes.c_int16),
        ('raw_value', ctypes.c_int16 * 1)]

class dbr_time_float(ctypes.Structure):
    dtype = numpy.float32
    scalar = ca_float
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('raw_value', ctypes.c_float * 1)]

class dbr_time_enum(ctypes.Structure):
    dtype = numpy.uint16
    scalar = ca_int
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('RISC_pad',  ctypes.c_int16),
        ('raw_value', ctypes.c_uint16 * 1)]

class dbr_time_char(ctypes.Structure):
    dtype = numpy.uint8
    scalar = ca_int
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('RISC_pad0', ctypes.c_int16),
        ('RISC_pad1', ctypes.c_uint8),
        ('raw_value', ctypes.c_uint8 * 1)]

class dbr_time_long(ctypes.Structure):
    dtype = numpy.int32
    scalar = ca_int
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('raw_value', ctypes.c_int32 * 1)]

class dbr_time_double(ctypes.Structure):
    dtype = numpy.float64
    scalar = ca_float
    copy_attributes = copy_attributes_time
    _fields_ = [
        ('status',    ctypes.c_int16),
        ('severity',  ctypes.c_int16),
        ('raw_stamp', ca_timestamp),
        ('RISC_pad',  ctypes.c_int32),
        ('raw_value', ctypes.c_double * 1)]

# DBR types with full control and graphical fields

class dbr_ctrl_short(ctypes.Structure):
    dtype = numpy.int16
    scalar = ca_int
    copy_attributes = copy_attributes_ctrl
    _fields_ = [
        ('status',              ctypes.c_int16),
        ('severity',            ctypes.c_int16),
        ('units',               ctypes.c_char * MAX_UNITS_SIZE),
        ('upper_disp_limit',    ctypes.c_int16),
        ('lower_disp_limit',    ctypes.c_int16),
        ('upper_alarm_limit',   ctypes.c_int16),
        ('upper_warning_limit', ctypes.c_int16),
        ('lower_warning_limit', ctypes.c_int16),
        ('lower_alarm_limit',   ctypes.c_int16),
        ('upper_ctrl_limit',    ctypes.c_int16),
        ('lower_ctrl_limit',    ctypes.c_int16),
        ('raw_value',           ctypes.c_int16 * 1)]

class dbr_ctrl_float(ctypes.Structure):
    dtype = numpy.float32
    scalar = ca_float
    copy_attributes = copy_attributes_ctrl
    _fields_ = [
        ('status',              ctypes.c_int16),
        ('severity',            ctypes.c_int16),
        ('precision',           ctypes.c_int16),
        ('RISC_pad',            ctypes.c_int16),
        ('units',               ctypes.c_char * MAX_UNITS_SIZE),
        ('upper_disp_limit',    ctypes.c_float),
        ('lower_disp_limit',    ctypes.c_float),
        ('upper_alarm_limit',   ctypes.c_float),
        ('upper_warning_limit', ctypes.c_float),
        ('lower_warning_limit', ctypes.c_float),
        ('lower_alarm_limit',   ctypes.c_float),
        ('upper_ctrl_limit',    ctypes.c_float),
        ('lower_ctrl_limit',    ctypes.c_float),
        ('raw_value',           ctypes.c_float * 1)]

class dbr_ctrl_enum(ctypes.Structure):
    dtype = numpy.uint16
    scalar = ca_int
    _fields_ = [
        ('status',   ctypes.c_int16),
        ('severity', ctypes.c_int16),
        ('no_str',   ctypes.c_int16),
        ('raw_strs', (ctypes.c_char * MAX_ENUM_STRING_SIZE) * MAX_ENUM_STATES),
        ('raw_value', ctypes.c_uint16 * 1)]

    def copy_attributes(self, other):
        other.status = self.status
        other.severity = self.severity
        other.enums = [
            py23.decode(ctypes.string_at(s))
            for s in self.raw_strs[:self.no_str]]

class dbr_ctrl_char(ctypes.Structure):
    dtype = numpy.uint8
    scalar = ca_int
    copy_attributes = copy_attributes_ctrl
    _fields_ = [
        ('status',              ctypes.c_int16),
        ('severity',            ctypes.c_int16),
        ('units',               ctypes.c_char * MAX_UNITS_SIZE),
        ('upper_disp_limit',    ctypes.c_uint8),
        ('lower_disp_limit',    ctypes.c_uint8),
        ('upper_alarm_limit',   ctypes.c_uint8),
        ('upper_warning_limit', ctypes.c_uint8),
        ('lower_warning_limit', ctypes.c_uint8),
        ('lower_alarm_limit',   ctypes.c_uint8),
        ('upper_ctrl_limit',    ctypes.c_uint8),
        ('lower_ctrl_limit',    ctypes.c_uint8),
        ('RISC_pad',            ctypes.c_uint8),
        ('raw_value',           ctypes.c_uint8 * 1)]

class dbr_ctrl_long(ctypes.Structure):
    dtype = numpy.int32
    scalar = ca_int
    copy_attributes = copy_attributes_ctrl
    _fields_ = [
        ('status',              ctypes.c_int16),
        ('severity',            ctypes.c_int16),
        ('units',               ctypes.c_char * MAX_UNITS_SIZE),
        ('upper_disp_limit',    ctypes.c_int32),
        ('lower_disp_limit',    ctypes.c_int32),
        ('upper_alarm_limit',   ctypes.c_int32),
        ('upper_warning_limit', ctypes.c_int32),
        ('lower_warning_limit', ctypes.c_int32),
        ('lower_alarm_limit',   ctypes.c_int32),
        ('upper_ctrl_limit',    ctypes.c_int32),
        ('lower_ctrl_limit',    ctypes.c_int32),
        ('raw_value',           ctypes.c_int32 * 1)]

class dbr_ctrl_double(ctypes.Structure):
    dtype = numpy.float64
    scalar = ca_float
    copy_attributes = copy_attributes_ctrl
    _fields_ = [
        ('status',              ctypes.c_int16),
        ('severity',            ctypes.c_int16),
        ('precision',           ctypes.c_int16),
        ('RISC_pad0',           ctypes.c_int16),
        ('units',               ctypes.c_char * MAX_UNITS_SIZE),
        ('upper_disp_limit',    ctypes.c_double),
        ('lower_disp_limit',    ctypes.c_double),
        ('upper_alarm_limit',   ctypes.c_double),
        ('upper_warning_limit', ctypes.c_double),
        ('lower_warning_limit', ctypes.c_double),
        ('lower_alarm_limit',   ctypes.c_double),
        ('upper_ctrl_limit',    ctypes.c_double),
        ('lower_ctrl_limit',    ctypes.c_double),
        ('raw_value',           ctypes.c_double * 1)]


class dbr_stsack_string(ctypes.Structure):
    dtype = str_dtype
    _fields_ = [
        ('status',              ctypes.c_int16),
        ('severity',            ctypes.c_int16),
        ('ackt',                ctypes.c_int16),
        ('acks',                ctypes.c_int16),
        ('raw_value',           (ctypes.c_byte * MAX_STRING_SIZE) * 1)]
    def copy_attributes(self, other):
        other.status = self.status
        other.severity = self.severity
        other.ackt = self.ackt
        other.acks = self.acks


# DBR request codes.  These correspond precisely to the types above, as
# identified in the DbrCodeToType lookup table below.
DBR_STRING = 0
DBR_SHORT = 1
DBR_FLOAT = 2
DBR_ENUM = 3
DBR_CHAR = 4
DBR_LONG = 5
DBR_DOUBLE = 6

DBR_TIME_STRING = 14
DBR_TIME_SHORT = 15
DBR_TIME_FLOAT = 16
DBR_TIME_ENUM = 17
DBR_TIME_CHAR = 18
DBR_TIME_LONG = 19
DBR_TIME_DOUBLE = 20

DBR_CTRL_SHORT = 29
DBR_CTRL_FLOAT = 30
DBR_CTRL_ENUM = 31
DBR_CTRL_CHAR = 32
DBR_CTRL_LONG = 33
DBR_CTRL_DOUBLE = 34

DBR_PUT_ACKT = 35       # Configure global alarm acknowledgement
DBR_PUT_ACKS = 36       # Acknowledge global alarm
DBR_STSACK_STRING = 37
DBR_CLASS_NAME = 38

# Special value for DBR_CHAR as str special processing.
DBR_ENUM_STR = 996
DBR_CHAR_BYTES = 997
DBR_CHAR_UNICODE = 998
DBR_CHAR_STR = 999


# Lookup table to convert supported DBR type codes into the corresponding DBR
# datatype.
DbrCodeToType = {
    DBR_STRING : dbr_string,
    DBR_SHORT : dbr_short,
    DBR_FLOAT : dbr_float,
    DBR_ENUM : dbr_enum,
    DBR_CHAR : dbr_char,
    DBR_LONG : dbr_long,
    DBR_DOUBLE : dbr_double,

    DBR_TIME_STRING : dbr_time_string,
    DBR_TIME_SHORT : dbr_time_short,
    DBR_TIME_FLOAT : dbr_time_float,
    DBR_TIME_ENUM : dbr_time_enum,
    DBR_TIME_CHAR : dbr_time_char,
    DBR_TIME_LONG : dbr_time_long,
    DBR_TIME_DOUBLE : dbr_time_double,

    DBR_CTRL_SHORT : dbr_ctrl_short,
    DBR_CTRL_FLOAT : dbr_ctrl_float,
    DBR_CTRL_ENUM : dbr_ctrl_enum,
    DBR_CTRL_CHAR : dbr_ctrl_char,
    DBR_CTRL_LONG : dbr_ctrl_long,
    DBR_CTRL_DOUBLE : dbr_ctrl_double,

    DBR_STSACK_STRING : dbr_stsack_string,
    DBR_CLASS_NAME : dbr_string,
}


# List of basic DBR types that we can process directly.
BasicDbrTypes = set([
    DBR_STRING,     DBR_SHORT,      DBR_FLOAT,      DBR_ENUM,
    DBR_CHAR,       DBR_LONG,       DBR_DOUBLE,
])


# Conversion from numpy character codes to DBR types.
NumpyCharCodeToDbr = {
    # The following type codes are supported directly:
    'b':    DBR_CHAR,       # byte   = int8
    'h':    DBR_SHORT,      # short  = int16
    'i':    DBR_LONG,       # intc   = int32
    'f':    DBR_FLOAT,      # single = float32
    'd':    DBR_DOUBLE,     # float_ = float64
    'S':    DBR_STRING,     # bytes_

    # The following type codes are weakly supported by pretending that
    # they're related types.
    '?':    DBR_CHAR,       # bool_
    'B':    DBR_CHAR,       # ubyte  = uint8
    'H':    DBR_SHORT,      # ushort = uint16
    'I':    DBR_LONG,       # uintc  = uint32

    # Unicode is supported by decoding from DBR_STRING
    'U':    DBR_STRING,     # str => unicode

    # We translate machine native integers to DBR_LONG as EPICS has no support
    # for 64-bit integers, but not allowing int as an argument is too confusing.
    'l':    DBR_LONG,       # int_ => int32, truncate as necessary
    'L':    DBR_LONG,       # uint => int32

    # The following type codes are not supported at all:
    #   q   longlong        Q   ulonglong       g   longfloat
    #   F   csingle         D   complex_        G   clongfloat
    #   O   object_         V   void            p, P    pointer types
}



# Format codes for type_to_dbr function.
FORMAT_RAW = 0
FORMAT_TIME = 1
FORMAT_CTRL = 2

class InvalidDatatype(Exception):
    '''Invalid datatype requested.'''

def _datatype_to_dbr(datatype):
    '''Converts Python datatype into a dbrcode and numpy dtype if possible,
    otherwise raises appropriate exception.'''
    try:
        # Rely on numpy for generic datatype recognition and conversion together
        # with filtering through our array of acceptable types.
        return NumpyCharCodeToDbr[numpy.dtype(datatype).char]
    except Exception as error:
        py23.raise_from(
            InvalidDatatype(
                'Datatype "%s" not supported for channel access' % datatype),
            error)

def _type_to_dbrcode(datatype, format):
    '''Converts a datatype and format request to a dbr value, or raises an
    exception if this cannot be done.

    datatype can be either a DBF_XXXX value as returned by ca_field_type() or
    data type supported by numpy

    format can be one of
      - FORMAT_RAW: retrieve the raw format only
      - FORMAT_TIME: retrieve timestamp and alarm status data
      - FORMAT_CTRL: retrieve limit and control data
    '''
    if datatype not in BasicDbrTypes:
        if datatype in [DBR_CHAR_STR, DBR_CHAR_BYTES, DBR_CHAR_UNICODE]:
            datatype = DBR_CHAR     # Retrieve this type using char array
        elif datatype in [DBR_STSACK_STRING, DBR_CLASS_NAME]:
            return datatype         # format is meaningless in this case
        else:
            datatype = _datatype_to_dbr(datatype)

    # Now take account of the format
    if format == FORMAT_RAW:
        # Use the raw datatype
        return datatype
    elif format == FORMAT_TIME:
        # Return corresponding DBR_TIME_XXXX value
        return datatype + 14
    elif format == FORMAT_CTRL:
        if datatype == DBR_STRING:
            # There is no ctrl option for strings, so in this case provide
            # the richest format we have available.
            return DBR_TIME_STRING
        else:
            # Return corresponding DBR_CTRL_XXX value
            return datatype + 28
    else:
        raise InvalidDatatype('Format not recognised')


# Helper functions for string arrays used in _convert_str_{str,bytes} below.
def _make_strings(raw_dbr, count):
    p_raw_value = ctypes.pointer(raw_dbr.raw_value[0])
    return [ctypes.string_at(p_raw_value[n]) for n in range(count)]

def _string_array(strings, count, dtypechar):
    if strings:
        n = max(len(s) for s in strings)
    else:
        # Zero length array of strings!  Very rare, but possible, but numpy
        # won't allow a dtype of S0, so we need to fake up something
        n = 1
    result = ca_array((count,), dtype = '%s%d' % (dtypechar, n))
    for i, s in enumerate(strings):
        result[i] = s
    return result

def _string_at(raw_value, count):
    # Need string_at() twice to ensure string is size limited *and* null
    # terminated.
    return ctypes.string_at(ctypes.string_at(raw_value, count))


# Conversion functions from raw_dbr to specified format.  These all take a
# raw_dbr and compute the appropriate selected Python type.  One of these
# conversion functions is selected internally with type_to_dbr below.

# Conversion from char array to strings
def _convert_char_str(raw_dbr, count):
    return ca_str(py23.decode(_string_at(raw_dbr.raw_value, count)))

# Conversion from char array to bytes strings
def _convert_char_bytes(raw_dbr, count):
    return ca_bytes(_string_at(raw_dbr.raw_value, count))

# Conversion from char array to unicode strings
def _convert_char_unicode(raw_dbr, count):
    return ca_unicode(_string_at(raw_dbr.raw_value, count).decode('UTF-8'))


# Arrays of standard strings.
def _convert_str_str(raw_dbr, count):
    return ca_str(py23.decode(_make_strings(raw_dbr, count)[0]))
def _convert_str_str_array(raw_dbr, count):
    strings = [py23.decode(s) for s in _make_strings(raw_dbr, count)]
    return _string_array(strings, count, str_char_code)

# Arrays of bytes strings.
def _convert_str_bytes(raw_dbr, count):
    return ca_bytes(_make_strings(raw_dbr, count)[0])
def _convert_str_bytes_array(raw_dbr, count):
    return _string_array(_make_strings(raw_dbr, count), count, 'S')

# Arrays of unicode strings.
def _convert_str_unicode(raw_dbr, count):
    return ca_str(_make_strings(raw_dbr, count)[0].decode('UTF-8'))
def _convert_str_unicode_array(raw_dbr, count):
    strings = [s.decode('UTF-8') for s in _make_strings(raw_dbr, count)]
    return _string_array(strings, count, 'U')


# For everything that isn't a string we either return a scalar or a ca_array
def _convert_other(raw_dbr, count):
    # Single elements are always returned as scalars.
    return raw_dbr.scalar(raw_dbr.raw_value[0])
def _convert_other_array(raw_dbr, count):
    # Build a fresh ca_array to receive a copy of the raw data in the dbr.
    # We have to take a copy, because the dbr is transient, and it is
    # helpful to use a numpy array as a container, because of the support it
    # provides.  It is essential that the dtype correctly matches the memory
    # layout of the raw dbr, and of course that the count is accurate.
    result = ca_array(shape = (count,), dtype = raw_dbr.dtype)
    ctypes.memmove(result.ctypes.data, raw_dbr.raw_value, result.nbytes)
    return result


def type_to_dbr(channel, datatype, format):
    '''Converts data request into the appropriate dbr code and conversion.  The
    channel must be ready so that its field type can be interrogated.  Returns
    dbr code together with conversion function for transforming dbr values of
    that type back into Python values.'''

    name = channel.name
    if datatype is None:
        # Use natural channel data type if no type specified.
        datatype = cadef.ca_field_type(channel)
        # A special hack for char arrays with PV name ending in $ -- these are
        # handled specially by EPICS as long strings.
        if datatype == DBR_CHAR and name[-1] == '$':
            datatype = DBR_CHAR_STR
    elif datatype == DBR_ENUM_STR:
        # A similar hack: for DBR_ENUM_STR use natural channel data type except
        # for enums which are fetched as strings.
        datatype = cadef.ca_field_type(channel)
        if datatype == DBR_ENUM:
            datatype = DBR_STRING

    # Prepare as much beforehand for conversion.
    dbrcode = _type_to_dbrcode(datatype, format)
    dbr_type = DbrCodeToType[dbrcode]
    dtype = dbr_type.dtype
    element_count = cadef.ca_element_count(channel)

    # Determine precisely which conversion from dbr to Python is required: all
    # the options for strings add a lot of complexity, ordinary numeric values
    # are all handled uniformly.
    if dtype is numpy.uint8 and datatype == DBR_CHAR_STR:
        # Conversion from char array to strings
        convert = _convert_char_str
    elif dtype is numpy.uint8 and datatype == DBR_CHAR_BYTES:
        # Conversion from char array to bytes strings
        convert = _convert_char_bytes
    elif dtype is numpy.uint8 and datatype == DBR_CHAR_UNICODE:
        # Conversion from char array to unicode strings
        convert = _convert_char_unicode
    else:
        if dtype is str_dtype:
            # String arrays, either unicode or normal.
            if isinstance(datatype, type) and issubclass(datatype, bytes):
                convert = (_convert_str_bytes, _convert_str_bytes_array)
            elif isinstance(datatype, type) and issubclass(datatype, unicode):
                convert = (_convert_str_unicode, _convert_str_unicode_array)
            else:
                convert = (_convert_str_str, _convert_str_str_array)
        else:
            convert = (_convert_other, _convert_other_array)

        # The conversion to scalar or array is determined by the original
        # element count of the underlying data source.
        if element_count == 1:
            convert = convert[0]
        else:
            convert = convert[1]


    # We return this function to perform conversion from dbr to Python value.
    def dbr_to_value(raw_dbr, dbrcode_in, count):
        # If the dbrcode has changed (this really shouldn't happen) then we've
        # got a problem!  If this does happen I'll need to handle this better,
        # as this is a pretty poor place to raise an exception.
        assert dbrcode_in == dbrcode, 'Oops, I didn\'t expect CA to do that'

        # Reinterpret the raw_dbr as a pointer to the appropriate structure as
        # identified by the given dbrcode.  We can then cast the raw_dbr
        # structure into an instance of this dbr: the data we want is then
        # available in the .raw_dbr field of this structure.
        raw_dbr = ctypes.cast(raw_dbr, ctypes.POINTER(dbr_type))[0]
        result = convert(raw_dbr, count)

        # Finally copy across any attributes together with the pv name and a
        # success indicator.
        raw_dbr.copy_attributes(result)
        result.name = name
        result.ok = True
        result.element_count = element_count
        result.datatype = datatype
        return result

    return dbrcode, dbr_to_value


# -----------------------------------------------------------------------------
# From Python value to DBR encoding, used by caput()

def _datatype_to_dtype(datatype):
    '''Converts any user specified datatype into dbrcode and dtype.'''
    if datatype not in BasicDbrTypes:
        datatype = _datatype_to_dbr(datatype)
    return datatype, DbrCodeToType[datatype].dtype


def _require_value(value, dtype):
    '''Use numpy to convert value into specified target type ready for transport
    over channel access.'''
    result = numpy.require(value, requirements = 'C', dtype = dtype)
    if result.shape == ():
        result.shape = (1,)
    assert result.ndim == 1, 'Can\'t put multidimensional arrays!'
    return result


def value_to_dbr(channel, datatype, value):
    '''Takes an ordinary Python value and converts it into a value in dbr
    format suitable for sending to channel access.  Returns the target
    datatype and the number of elements together with a pointer to the raw
    data and (for lifetime management) the object containing the data.'''

    # If no datatype specified then use the target datatype.
    if datatype is None:
        if isinstance(value, (str, bytes, unicode)):
            # Give strings with no datatype special treatment, let the IOC do
            # the decoding.  It's safer this way.
            datatype = DBR_STRING
        else:
            datatype = cadef.ca_field_type(channel)
            if datatype == DBR_CHAR and channel.name[-1] == '$':
                # Treat char arrays with name ending in $ as long strings.
                datatype = DBR_CHAR_STR

    if datatype == DBR_CHAR_STR:
        # Char arrays as strings need special treatment.
        count = cadef.ca_element_count(channel)
        try:
            result = _require_value(value, 'S%d' % count)
        except UnicodeEncodeError:
            # Unicode needs to be encoded
            result = _require_value(value.encode('UTF-8'), 'S%d' % count)
        assert result.shape[0] == 1, \
            'Can\'t put array of strings as char array'
        return DBR_CHAR, count, result.ctypes.data, result
    elif datatype in [DBR_PUT_ACKT, DBR_PUT_ACKS]:
        # For DBR_PUT_ACKT and DBR_PUT_ACKS we return an integer
        value = ctypes.c_int32(value)
        return datatype, 1, ctypes.byref(value), value
    else:
        # For all other types compute the appropriate transport type
        dbrcode, dtype = _datatype_to_dtype(datatype)
        if dbrcode is DBR_STRING:
            try:
                # We'll let numpy do most of the heavy lifting.
                result = _require_value(value, str_dtype)
            except UnicodeEncodeError:
                # Whoops, looks like unicode, need to encode each string.
                value = _require_value(value, None)
                result = numpy.empty(value.shape, str_dtype)
                for n, s in enumerate(value):
                    result[n] = s.encode('UTF-8')
        else:
            # Numpy can do all the conversion for all the remaining data types.
            result = _require_value(value, dtype)

        return dbrcode, len(result), result.ctypes.data, result
